# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

import braket.ir.jaqcd as ir
from braket.circuits import Circuit, Observable, ResultType
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.result_type import ObservableParameterResultType
from braket.circuits.result_types import ObservableResultType
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)

testdata = [
    (ResultType.StateVector, "state_vector", ir.StateVector, {}, {}),
    (ResultType.DensityMatrix, "density_matrix", ir.DensityMatrix, {}, {}),
    (
        ResultType.DensityMatrix,
        "density_matrix",
        ir.DensityMatrix,
        {"target": [0, 1]},
        {"targets": [0, 1]},
    ),
    (ResultType.Amplitude, "amplitude", ir.Amplitude, {"state": ["0"]}, {"states": ["0"]}),
    (
        ResultType.Probability,
        "probability",
        ir.Probability,
        {"target": [0, 1]},
        {"targets": [0, 1]},
    ),
    (
        ResultType.Probability,
        "probability",
        ir.Probability,
        {"target": None},
        {},
    ),
    (
        ResultType.Expectation,
        "expectation",
        ir.Expectation,
        {"observable": Observable.Z(), "target": [0]},
        {"observable": ["z"], "targets": [0]},
    ),
    (
        ResultType.Expectation,
        "expectation",
        ir.Expectation,
        {"observable": Observable.Z() @ Observable.I() @ Observable.X(), "target": [0, 1, 2]},
        {"observable": ["z", "i", "x"], "targets": [0, 1, 2]},
    ),
    (
        ResultType.Expectation,
        "expectation",
        ir.Expectation,
        {"observable": Observable.Hermitian(matrix=Observable.I().to_matrix()), "target": [0]},
        {"observable": [[[[1.0, 0], [0, 0]], [[0, 0], [1.0, 0]]]], "targets": [0]},
    ),
    (
        ResultType.Expectation,
        "expectation",
        ir.Expectation,
        {"observable": Observable.Hermitian(matrix=Observable.I().to_matrix()), "target": None},
        {"observable": [[[[1.0, 0], [0, 0]], [[0, 0], [1.0, 0]]]]},
    ),
    (
        ResultType.Sample,
        "sample",
        ir.Sample,
        {"observable": Observable.Z(), "target": [0]},
        {"observable": ["z"], "targets": [0]},
    ),
    (
        ResultType.Sample,
        "sample",
        ir.Sample,
        {"observable": Observable.Z(), "target": None},
        {"observable": ["z"]},
    ),
    (
        ResultType.Variance,
        "variance",
        ir.Variance,
        {"observable": Observable.Z(), "target": [0]},
        {"observable": ["z"], "targets": [0]},
    ),
    (
        ResultType.Variance,
        "variance",
        ir.Variance,
        {"observable": Observable.Z(), "target": None},
        {"observable": ["z"]},
    ),
    (
        ResultType.AdjointGradient,
        "adjoint_gradient",
        ir.AdjointGradient,
        {"observable": Observable.Z(), "target": None, "parameters": None},
        {"observable": ["z"]},
    ),
    (
        ResultType.AdjointGradient,
        "adjoint_gradient",
        ir.AdjointGradient,
        {"observable": Observable.Z(), "target": [0], "parameters": ["alpha", "beta"]},
        {"observable": ["z"], "targets": [0], "parameters": ["alpha", FreeParameter("beta")]},
    ),
    (
        ResultType.AdjointGradient,
        "adjoint_gradient",
        ir.AdjointGradient,
        {
            "observable": Observable.Hermitian(matrix=Observable.Z().to_matrix()),
            "target": [0],
            "parameters": ["alpha", "beta"],
        },
        {
            "observable": [[[[1.0, 0], [0, 0]], [[0, 0], [-1.0, 0]]]],
            "targets": [0],
            "parameters": ["alpha", "beta"],
        },
    ),
]


@pytest.mark.parametrize("testclass,subroutine_name,irclass,input,ir_input", testdata)
def test_ir_result_level(testclass, subroutine_name, irclass, input, ir_input):
    if testclass == ResultType.AdjointGradient:
        jaqcd_not_implemented = "to_jaqcd has not been implemented yet."
        with pytest.raises(NotImplementedError, match=jaqcd_not_implemented):
            testclass(**input).to_ir()
    else:
        expected = irclass(**ir_input)
        actual = testclass(**input).to_ir()
        assert actual == expected


@pytest.mark.parametrize(
    "result_type, serialization_properties, expected_ir",
    [
        (
            ResultType.Expectation(Observable.I(), target=0),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result expectation i(q[0])",
        ),
        (
            ResultType.Expectation(Observable.I()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result expectation i all",
        ),
        (
            ResultType.StateVector(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result state_vector",
        ),
        (
            ResultType.DensityMatrix(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result density_matrix all",
        ),
        (
            ResultType.DensityMatrix([0, 2]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result density_matrix q[0], q[2]",
        ),
        (
            ResultType.DensityMatrix(0),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket result density_matrix $0",
        ),
        (
            ResultType.Amplitude(["01", "10"]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            '#pragma braket result amplitude "01", "10"',
        ),
        (
            ResultType.Probability(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result probability all",
        ),
        (
            ResultType.Probability([0, 2]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result probability q[0], q[2]",
        ),
        (
            ResultType.Probability(0),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket result probability $0",
        ),
        (
            ResultType.Sample(Observable.I(), target=0),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result sample i(q[0])",
        ),
        (
            ResultType.Sample(Observable.I()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result sample i all",
        ),
        (
            ResultType.Variance(Observable.I(), target=0),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result variance i(q[0])",
        ),
        (
            ResultType.Variance(Observable.I()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result variance i all",
        ),
        (
            ResultType.AdjointGradient(Observable.I(), target=0, parameters=["alpha", "beta"]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(i(q[0])) alpha, beta",
        ),
        (
            ResultType.AdjointGradient(Observable.I(), target=0, parameters=["alpha"]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(i(q[0])) alpha",
        ),
        (
            ResultType.AdjointGradient(
                Observable.H() @ Observable.I(),
                target=[0, 1],
                parameters=[FreeParameter("alpha"), "beta", FreeParameter("gamma")],
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(h(q[0]) @ i(q[1])) "
            "alpha, beta, gamma",
        ),
        (
            ResultType.AdjointGradient(
                Observable.H() @ Observable.I() + 2 * Observable.Z(),
                target=[[0, 1], [2]],
                parameters=[FreeParameter("alpha"), "beta", FreeParameter("gamma")],
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(h(q[0]) @ i(q[1]) + 2 * z(q[2])) "
            "alpha, beta, gamma",
        ),
        (
            ResultType.AdjointGradient(Observable.I(), target=0, parameters=[]),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(i(q[0])) all",
        ),
        (
            ResultType.AdjointGradient(
                Observable.X() @ Observable.Y(), target=[0, 1], parameters=[]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(x(q[0]) @ y(q[1])) all",
        ),
        (
            ResultType.AdjointGradient(
                Observable.Hermitian(matrix=Observable.I().to_matrix()), target=0, parameters=[]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket result adjoint_gradient expectation(hermitian([[1+0im, 0im], "
            "[0im, 1+0im]]) q[0]) all",
        ),
    ],
)
def test_result_to_ir_openqasm(result_type, serialization_properties, expected_ir):
    assert (
        result_type.to_ir(IRType.OPENQASM, serialization_properties=serialization_properties)
        == expected_ir
    )


@pytest.mark.parametrize("testclass,subroutine_name,irclass,input,ir_input", testdata)
def test_result_subroutine(testclass, subroutine_name, irclass, input, ir_input):
    subroutine = getattr(Circuit(), subroutine_name)
    assert subroutine(**input) == Circuit(testclass(**input))


@pytest.mark.parametrize("testclass,subroutine_name,irclass,input,ir_input", testdata)
def test_result_equality(testclass, subroutine_name, irclass, input, ir_input):
    a1 = testclass(**input)
    a2 = a1.copy()
    assert a1 == a2
    assert a1 is not a2


# Amplitude


@pytest.mark.parametrize(
    "state",
    ((["2", "11"]), ([1, 0]), ([0.1, 0]), ("-0", "1"), (["", ""]), (None), ([None, None]), ("10")),
)
def test_amplitude_init_invalid_state_value_error(state):
    with pytest.raises(ValueError):
        ResultType.Amplitude(state=state)


def test_amplitude_equality():
    a1 = ResultType.Amplitude(state=["0", "1"])
    a2 = ResultType.Amplitude(state=["0", "1"])
    a3 = ResultType.Amplitude(state=["01", "11", "10"])
    a4 = "hi"
    assert a1 == a2
    assert a1 != a3
    assert a1 != a4


# Probability


def test_probability_equality():
    a1 = ResultType.Probability([0, 1])
    a2 = ResultType.Probability([0, 1])
    a3 = ResultType.Probability([0, 1, 2])
    a4 = "hi"
    assert a1 == a2
    assert a1 != a3
    assert a1 != a4


# Expectation


def test_expectation_parent_class():
    assert isinstance(
        ResultType.Expectation(observable=Observable.X(), target=0), ObservableResultType
    )


# Sample


def test_sample_parent_class():
    assert isinstance(ResultType.Sample(observable=Observable.X(), target=0), ObservableResultType)


# Variance


def test_variance_parent_class():
    assert isinstance(
        ResultType.Variance(observable=Observable.X(), target=0), ObservableResultType
    )


# AdjointGradient


def test_adjoint_gradient_parent_class():
    assert isinstance(
        ResultType.AdjointGradient(
            observable=Observable.X(), target=0, parameters=["alpha", FreeParameter("beta")]
        ),
        ObservableParameterResultType,
    )


@pytest.mark.parametrize(
    "target",
    (
        [[0], [0, 1], [2]],
        [[0, 1], [0, 1]],
    ),
)
def test_incorrect_target_adjoint_gradient(target):
    match = (
        "Sum observable's target shape must be a nested list "
        "where each term's target length is equal to the observable term's qubits count."
    )
    with pytest.raises(ValueError, match=match):
        ResultType.AdjointGradient(
            2 * Observable.Z() + 3 * Observable.X() @ Observable.Y(),
            target,
        )



from botocore import awsrequest

import braket._schemas as braket_schemas
import braket._sdk as braket_sdk


def test_default_user_agent(aws_session):
    braket_agents = [
        f"BraketSdk/{braket_sdk.__version__}",
        f"BraketSchemas/{braket_schemas.__version__}",
        "NotebookInstance/",
    ]

    def assert_in_user_agent(request: awsrequest.AWSPreparedRequest, **kwargs):
        user_agent = request.headers.get("User-Agent")
        for agent in braket_agents:
            assert bytes(agent, "utf8") in user_agent

    aws_session.braket_client.meta.events.register("before-send.braket", assert_in_user_agent)

    aws_session.search_devices()


def test_add_user_agent(aws_session):
    aws_session.add_braket_user_agent("foo/1.0")

    def assert_in_user_agent(request: awsrequest.AWSPreparedRequest, **kwargs):
        user_agent = request.headers.get("User-Agent")
        assert b"foo/1.0" in user_agent

    aws_session.braket_client.meta.events.register("before-send.braket", assert_in_user_agent)

    aws_session.search_devices()



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import math

import numpy as np
import numpy.testing as npt
import pytest

from braket.circuits import Gate, Observable
from braket.circuits.observables import observable_from_ir
from braket.circuits.quantum_operator_helpers import get_pauli_eigenvalues
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)

testdata = [
    (Observable.I(), Gate.I(), ["i"], (), np.array([1, 1])),
    (Observable.X(), Gate.X(), ["x"], tuple([Gate.H()]), get_pauli_eigenvalues(1)),
    (
        Observable.Y(),
        Gate.Y(),
        ["y"],
        tuple([Gate.Z(), Gate.S(), Gate.H()]),
        get_pauli_eigenvalues(1),
    ),
    (Observable.Z(), Gate.Z(), ["z"], (), get_pauli_eigenvalues(1)),
    (Observable.H(), Gate.H(), ["h"], tuple([Gate.Ry(-math.pi / 4)]), get_pauli_eigenvalues(1)),
]

invalid_hermitian_matrices = [
    (np.array([[1]])),
    (np.array([1])),
    (np.array([0, 1, 2])),
    (np.array([[0, 1], [1, 2], [3, 4]])),
    (np.array([[0, 1, 2], [2, 3]], dtype=object)),
    (np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])),
    (Gate.T().to_matrix()),
]


@pytest.mark.parametrize(
    "testobject,gateobject,expected_ir,basis_rotation_gates,eigenvalues", testdata
)
def test_to_ir(testobject, gateobject, expected_ir, basis_rotation_gates, eigenvalues):
    expected = expected_ir
    actual = testobject.to_ir()
    assert actual == expected


@pytest.mark.parametrize(
    "observable, serialization_properties, target, expected_ir",
    [
        (
            Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "i(q[3])",
        ),
        (
            Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "i($3)",
        ),
        (
            Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "i all",
        ),
        (
            Observable.X(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "x(q[3])",
        ),
        (
            Observable.X(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "x($3)",
        ),
        (
            Observable.X(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "x all",
        ),
        (
            Observable.Y(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "y(q[3])",
        ),
        (
            Observable.Y(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "y($3)",
        ),
        (
            Observable.Y(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "y all",
        ),
        (
            Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "z(q[3])",
        ),
        (
            Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "z($3)",
        ),
        (
            Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "z all",
        ),
        (
            Observable.H(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "h(q[3])",
        ),
        (
            Observable.H(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "h($3)",
        ),
        (
            Observable.H(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "h all",
        ),
        (
            Observable.Hermitian(np.eye(4)),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [1, 2],
            "hermitian([[1+0im, 0im, 0im, 0im], [0im, 1+0im, 0im, 0im], "
            "[0im, 0im, 1+0im, 0im], [0im, 0im, 0im, 1+0im]]) q[1], q[2]",
        ),
        (
            Observable.Hermitian(np.eye(4)),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [1, 2],
            "hermitian([[1+0im, 0im, 0im, 0im], [0im, 1+0im, 0im, 0im], "
            "[0im, 0im, 1+0im, 0im], [0im, 0im, 0im, 1+0im]]) $1, $2",
        ),
        (
            Observable.Hermitian(np.eye(2)),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            None,
            "hermitian([[1+0im, 0im], [0im, 1+0im]]) all",
        ),
        (
            Observable.H() @ Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 0],
            "h(q[3]) @ z(q[0])",
        ),
        (
            Observable.H() @ Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 0],
            "h($3) @ z($0)",
        ),
        (
            Observable.H() @ Observable.Z() @ Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 0, 1],
            "h(q[3]) @ z(q[0]) @ i(q[1])",
        ),
        (
            Observable.H() @ Observable.Z() @ Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 0, 1],
            "h($3) @ z($0) @ i($1)",
        ),
        (
            Observable.Hermitian(np.eye(4)) @ Observable.I(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 0, 1],
            "hermitian([[1+0im, 0im, 0im, 0im], [0im, 1+0im, 0im, 0im], "
            "[0im, 0im, 1+0im, 0im], [0im, 0im, 0im, 1+0im]]) q[3], q[0]"
            " @ i(q[1])",
        ),
        (
            Observable.I() @ Observable.Hermitian(np.eye(4)),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 0, 1],
            "i($3) @ "
            "hermitian([[1+0im, 0im, 0im, 0im], [0im, 1+0im, 0im, 0im], "
            "[0im, 0im, 1+0im, 0im], [0im, 0im, 0im, 1+0im]]) $0, $1",
        ),
        (
            (2 * Observable.Z()) @ (3 * Observable.H()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 3],
            "6 * z($3) @ h($3)",
        ),
        (
            (2 * Observable.Z()) @ (3 * Observable.H()) @ (2 * Observable.Y()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 3, 1],
            "12 * z($3) @ h($3) @ y($1)",
        ),
        (
            3 * (2 * Observable.Z()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "6 * z($3)",
        ),
        (
            (2 * Observable.I()) @ (2 * Observable.Hermitian(np.eye(4))),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 0, 1],
            "4 * i($3) @ "
            "hermitian([[1+0im, 0im, 0im, 0im], [0im, 1+0im, 0im, 0im], "
            "[0im, 0im, 1+0im, 0im], [0im, 0im, 0im, 1+0im]]) $0, $1",
        ),
        (
            Observable.Z() + 2 * Observable.H(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[3], [4]],
            "z($3) + 2 * h($4)",
        ),
        (
            3 * (Observable.H() + 2 * Observable.X()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[3], [0]],
            "3 * h($3) + 6 * x($0)",
        ),
        (
            3 * (Observable.H() + 2 * Observable.H()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[3], [3]],
            "3 * h($3) + 6 * h($3)",
        ),
        (
            3 * (Observable.H() + 2 * Observable.H()),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[3], [5]],
            "3 * h($3) + 6 * h($5)",
        ),
        (
            (2 * Observable.Y()) @ (3 * Observable.I()) + 0.75 * Observable.Y() @ Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[0, 1], [0, 1]],
            "6 * y($0) @ i($1) + 0.75 * y($0) @ z($1)",
        ),
        (
            (-2 * Observable.Y()) @ (3 * Observable.I()) + -0.75 * Observable.Y() @ Observable.Z(),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[0, 1], [0, 1]],
            "-6 * y($0) @ i($1) - 0.75 * y($0) @ z($1)",
        ),
        (
            4 * (2 * Observable.Z() + 2 * (3 * Observable.X() @ (2 * Observable.Y()))),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [[0], [1, 2]],
            "8 * z($0) + 48 * x($1) @ y($2)",
        ),
    ],
)
def test_observables_to_ir_openqasm(observable, serialization_properties, target, expected_ir):
    assert (
        observable.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )
        == expected_ir
    )


@pytest.mark.parametrize(
    "observable",
    [
        2 * Observable.H(),
        3 * Observable.Z(),
        2 * Observable.I(),
        3 * Observable.X(),
        2 * Observable.Y(),
        2 * Observable.Hermitian(matrix=np.array([[0, 1], [1, 0]])),
        2 * Observable.TensorProduct([Observable.Z(), Observable.H()]),
    ],
)
def test_observable_coef_jaqcd(observable):
    coef_not_supported_with_jaqcd = "Observable coefficients not supported with Jaqcd"
    with pytest.raises(ValueError, match=coef_not_supported_with_jaqcd):
        observable.to_ir(target=0, ir_type=IRType.JAQCD)


@pytest.mark.parametrize(
    "expression, observable",
    [
        ([], Observable.X()),
        ([2], Observable.Y()),
        ([2, "invalid_str"], Observable.Z()),
        ([2.0], Observable.Hermitian(matrix=np.array([[0, 1], [1, 0]]))),
        ([2], Observable.Sum([Observable.X() + Observable.Y()])),
        ([2], Observable.Y() + 0.75 * Observable.Y() @ Observable.Z()),
    ],
)
def test_invalid_scalar_multiplication(expression, observable):
    with pytest.raises(TypeError, match="Observable coefficients must be numbers."):
        expression * observable


@pytest.mark.parametrize(
    "observable, matrix",
    [
        (
            (-3 * Observable.H()).to_matrix(),
            np.array(
                [[-2.12132034 + 0.0j, -2.12132034 + 0.0j], [-2.12132034 + 0.0j, 2.12132034 - 0.0j]]
            ),
        ),
        (
            (3 * Observable.Z()).to_matrix(),
            np.array([[3.0 + 0.0j, 0.0 + 0.0j], [0.0 + 0.0j, -3.0 + 0.0j]]),
        ),
        (
            (2 * Observable.I()).to_matrix(),
            np.array([[2.0 + 0.0j, 0.0 + 0.0j], [0.0 + 0.0j, 2.0 + 0.0j]]),
        ),
        (
            (1.2 * Observable.X()).to_matrix(),
            np.array([[0.0 + 0.0j, 1.2 + 0.0j], [1.2 + 0.0j, 0.0 + 0.0j]]),
        ),
        (
            (1e-2 * Observable.Y()).to_matrix(),
            np.array([[0.0 + 0.0j, 0.0 - 0.01j], [0 + 0.01j, 0.0 + 0.0j]]),
        ),
        (
            (np.array(1.3) * Observable.Hermitian(matrix=np.array([[0, 1], [1, 0]]))).to_matrix(),
            np.array([[0.0 + 0.0j, 1.3 + 0.0j], [1.3 + 0.0j, 0.0 + 0.0j]]),
        ),
        (
            (2 * Observable.TensorProduct([Observable.Z(), Observable.H()])).to_matrix(),
            np.array(
                [
                    [1.41421356 + 0.0j, 1.41421356 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                    [1.41421356 + 0.0j, -1.41421356 + 0.0j, 0.0 + 0.0j, -0.0 + 0.0j],
                    [0.0 + 0.0j, 0.0 + 0.0j, -1.41421356 + 0.0j, -1.41421356 + 0.0j],
                    [0.0 + 0.0j, -0.0 + 0.0j, -1.41421356 + 0.0j, 1.41421356 + 0.0j],
                ],
            ),
        ),
    ],
)
def test_valid_scaled_matrix(observable, matrix):
    npt.assert_allclose(observable, matrix)


@pytest.mark.parametrize(
    "observable, eigenvalue",
    [
        (-2 * Observable.I().eigenvalues, np.array([-2.0, -2.0])),
        (
            3e-2 * Observable.Hermitian(matrix=np.array([[0, 1], [1, 0]])).eigenvalues,
            np.array([-0.03, 0.03]),
        ),
    ],
)
def test_valid_scaled_eigenvalues(observable, eigenvalue):
    npt.assert_allclose(observable, eigenvalue)


@pytest.mark.parametrize(
    "testobject,gateobject,expected_ir,basis_rotation_gates,eigenvalues", testdata
)
def test_gate_equality(testobject, gateobject, expected_ir, basis_rotation_gates, eigenvalues):
    assert testobject.qubit_count == gateobject.qubit_count
    assert testobject.ascii_symbols == gateobject.ascii_symbols
    assert testobject.matrix_equivalence(gateobject)
    assert testobject.basis_rotation_gates == basis_rotation_gates
    assert np.allclose(testobject.eigenvalues, eigenvalues)


@pytest.mark.parametrize(
    "testobject,gateobject,expected_ir,basis_rotation_gates,eigenvalues", testdata
)
def test_basis_rotation_gates(
    testobject, gateobject, expected_ir, basis_rotation_gates, eigenvalues
):
    assert testobject.basis_rotation_gates == basis_rotation_gates


@pytest.mark.parametrize(
    "testobject,gateobject,expected_ir,basis_rotation_gates,eigenvalues", testdata
)
def test_eigenvalues(testobject, gateobject, expected_ir, basis_rotation_gates, eigenvalues):
    compare_eigenvalues(testobject, eigenvalues)


@pytest.mark.parametrize(
    "testobject,gateobject,expected_ir,basis_rotation_gates,eigenvalues", testdata
)
def test_observable_from_ir(testobject, gateobject, expected_ir, basis_rotation_gates, eigenvalues):
    assert testobject == observable_from_ir(expected_ir)


# Hermitian


@pytest.mark.parametrize("matrix", invalid_hermitian_matrices)
def test_hermitian_invalid_matrix(matrix):
    with pytest.raises(ValueError):
        Observable.Hermitian(matrix=matrix)


def test_hermitian_equality():
    matrix = Observable.H().to_matrix()
    a1 = Observable.Hermitian(matrix=matrix)
    a2 = Observable.Hermitian(matrix=matrix)
    a3 = Observable.Hermitian(matrix=Observable.I().to_matrix())
    a4 = "hi"
    assert a1 == a2
    assert a1 != a3
    assert a1 != a4


def test_hermitian_to_ir():
    matrix = Observable.I().to_matrix()
    obs = Observable.Hermitian(matrix=matrix)
    assert obs.to_ir() == [[[[1, 0], [0, 0]], [[0, 0], [1, 0]]]]


@pytest.mark.parametrize(
    "matrix,eigenvalues",
    [
        (np.array([[1.0, 0.0], [0.0, 1.0]]), np.array([1, 1])),
        (np.array([[0, -1j], [1j, 0]]), np.array([-1.0, 1.0])),
        (np.array([[1, 1 - 1j], [1 + 1j, -1]]), np.array([-np.sqrt(3), np.sqrt(3)])),
    ],
)
def test_hermitian_eigenvalues(matrix, eigenvalues):
    compare_eigenvalues(Observable.Hermitian(matrix=matrix), eigenvalues)


def test_flattened_tensor_product():
    observable_one = Observable.Z() @ Observable.Y()
    observable_two = Observable.X() @ Observable.H()
    actual = Observable.TensorProduct([observable_one, observable_two])
    expected = Observable.TensorProduct(
        [Observable.Z(), Observable.Y(), Observable.X(), Observable.H()]
    )
    assert expected == actual


@pytest.mark.parametrize(
    "matrix,basis_rotation_matrix",
    [
        (
            np.array([[0.0, 1.0], [1.0, 0.0]]),
            np.array([[-0.70710678, 0.70710678], [0.70710678, 0.70710678]]).conj().T,
        ),
        (
            np.array([[0, -1j], [1j, 0]]),
            np.array(
                [[-0.70710678 + 0.0j, -0.70710678 + 0.0j], [0.0 + 0.70710678j, 0.0 - 0.70710678j]]
            )
            .conj()
            .T,
        ),
        (
            np.array([[1, 1 - 1j], [1 + 1j, -1]]),
            np.array(
                [
                    [-0.45970084 - 0.0j, 0.62796303 - 0.62796303j],
                    [-0.88807383 - 0.0j, -0.32505758 + 0.32505758j],
                ]
            ),
        ),
    ],
)
def test_hermitian_basis_rotation_gates(matrix, basis_rotation_matrix):
    expected_unitary = Gate.Unitary(matrix=basis_rotation_matrix)
    actual_rotation_gates = Observable.Hermitian(matrix=matrix).basis_rotation_gates
    assert actual_rotation_gates == tuple([expected_unitary])
    assert expected_unitary.matrix_equivalence(actual_rotation_gates[0])


def test_observable_from_ir_hermitian_value_error():
    ir_observable = [[[[1.0, 0], [0, 1]], [[0.0, 1], [1, 0]]]]
    with pytest.raises(ValueError):
        observable_from_ir(ir_observable)


def test_observable_from_ir_hermitian():
    ir_observable = [[[[1, 0], [0, 0]], [[0, 0], [1, 0]]]]
    actual_observable = observable_from_ir(ir_observable)
    assert actual_observable == Observable.Hermitian(matrix=np.array([[1.0, 0.0], [0.0, 1.0]]))


def test_hermitian_str():
    assert (
        str(Observable.Hermitian(matrix=np.array([[1.0, 0.0], [0.0, 1.0]])))
        == "Hermitian('qubit_count': 1, 'matrix': [[1.+0.j 0.+0.j], [0.+0.j 1.+0.j]])"
    )


# TensorProduct


def test_tensor_product_to_ir():
    t = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    assert t.to_ir() == ["z", "i", "x"]
    assert t.qubit_count == 3
    assert t.ascii_symbols == tuple(["Z@I@X"] * 3)


def test_tensor_product_matmul_tensor():
    t1 = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    t2 = Observable.TensorProduct(
        [Observable.Hermitian(matrix=Observable.I().to_matrix()), Observable.Y()]
    )
    t3 = t1 @ t2
    assert t3.to_ir() == ["z", "i", "x", [[[1.0, 0], [0, 0]], [[0, 0], [1.0, 0]]], "y"]
    assert t3.qubit_count == 5
    assert t3.ascii_symbols == tuple(["Z@I@X@Hermitian@Y"] * 5)


def test_tensor_product_matmul_observable():
    t1 = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    o1 = Observable.I()
    t = t1 @ o1
    assert t.to_ir() == ["z", "i", "x", "i"]
    assert t.qubit_count == 4
    assert t.ascii_symbols == tuple(["Z@I@X@I"] * 4)


def test_tensor_product_eigenvalue_index_out_of_bounds():
    obs = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    with pytest.raises(ValueError):
        obs.eigenvalue(8)


def test_tensor_product_value_error():
    with pytest.raises(ValueError):
        Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()]) @ "a"


def test_tensor_product_rmatmul_observable():
    t1 = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    o1 = Observable.I()
    t = o1 @ t1
    assert t.to_ir() == ["i", "z", "i", "x"]
    assert t.qubit_count == 4
    assert t.ascii_symbols == tuple(["I@Z@I@X"] * 4)


@pytest.mark.parametrize(
    "observable,eigenvalues",
    [
        (Observable.X() @ Observable.Y(), np.array([1, -1, -1, 1])),
        (Observable.X() @ Observable.Y() @ Observable.Z(), np.array([1, -1, -1, 1, -1, 1, 1, -1])),
        (Observable.X() @ Observable.Y() @ Observable.I(), np.array([1, 1, -1, -1, -1, -1, 1, 1])),
        (
            Observable.X()
            @ Observable.Hermitian(
                np.array([[-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
            )
            @ Observable.Y(),
            np.array([-1, 1, -1, 1, 1, -1, 1, -1, 1, -1, 1, -1, -1, 1, -1, 1]),
        ),
    ],
)
def test_tensor_product_eigenvalues(observable, eigenvalues):
    compare_eigenvalues(observable, eigenvalues)
    # Test caching
    observable._factors = ()
    compare_eigenvalues(observable, eigenvalues)


@pytest.mark.parametrize(
    "observable,basis_rotation_gates",
    [
        (Observable.X() @ Observable.Y(), tuple([Gate.H(), Gate.Z(), Gate.S(), Gate.H()])),
        (
            Observable.X() @ Observable.Y() @ Observable.Z(),
            tuple([Gate.H(), Gate.Z(), Gate.S(), Gate.H()]),
        ),
        (
            Observable.X() @ Observable.Y() @ Observable.I(),
            tuple([Gate.H(), Gate.Z(), Gate.S(), Gate.H()]),
        ),
        (Observable.X() @ Observable.H(), tuple([Gate.H(), Gate.Ry(-np.pi / 4)])),
    ],
)
def test_tensor_product_basis_rotation_gates(observable, basis_rotation_gates):
    assert observable.basis_rotation_gates == basis_rotation_gates


def test_observable_from_ir_tensor_product():
    expected_observable = Observable.TensorProduct([Observable.Z(), Observable.I(), Observable.X()])
    actual_observable = observable_from_ir(["z", "i", "x"])
    assert expected_observable == actual_observable


def test_observable_from_ir_tensor_product_value_error():
    with pytest.raises(ValueError):
        observable_from_ir(["z", "i", "foo"])


def compare_eigenvalues(observable, expected):
    assert np.allclose(observable.eigenvalues, expected)
    assert np.allclose(
        np.array([observable.eigenvalue(i) for i in range(2**observable.qubit_count)]),
        expected,
    )


def test_sum_not_allowed_in_tensor_product():
    sum_not_allowed_in_tensor_product = "Sum observables not allowed in TensorProduct"
    with pytest.raises(TypeError, match=sum_not_allowed_in_tensor_product):
        Observable.TensorProduct([Observable.X() + Observable.Y()])


# Sum of observables


@pytest.mark.parametrize(
    "observable,basis_rotation_gates",
    [
        (Observable.X() + Observable.Y(), tuple([Gate.H(), Gate.Z(), Gate.S(), Gate.H()])),
    ],
)
def test_no_basis_rotation_support_for_sum(observable, basis_rotation_gates):
    no_basis_rotation_support_for_sum = "Basis rotation calculation not supported for Sum"
    with pytest.raises(NotImplementedError, match=no_basis_rotation_support_for_sum):
        observable.basis_rotation_gates


def test_no_eigenvalues_support_for_sum():
    no_eigen_value_support = "Eigenvalue calculation not supported for Sum"
    with pytest.raises(NotImplementedError, match=no_eigen_value_support):
        (Observable.X() + Observable.Y()).eigenvalues


def test_matrix_not_supported_for_sum():
    matrix_not_supported = "Matrix operation is not supported for Sum"
    with pytest.raises(NotImplementedError, match=matrix_not_supported):
        (Observable.X() + Observable.Y()).to_matrix()


def test_invalid_targets_config_for_sum_obs():
    observable, serialization_properties = (
        2 * Observable.X() @ Observable.Y() + 0.75 * Observable.Y() @ Observable.Z(),
        OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
    )
    target = [[0, 1]]

    target_len_mismatch_for_sum_terms = "Invalid target of length 1 for Sum with 2 terms"

    with pytest.raises(ValueError, match=target_len_mismatch_for_sum_terms):
        observable.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )


def test_sum_obs_str():
    assert (
        str(Observable.Sum([2 * Observable.X() + 3 * Observable.Y()]))
        == "Sum(X('qubit_count': 1), Y('qubit_count': 1))"
    )


def test_str_equality_sum_obs():
    t1 = Observable.Sum([2 * Observable.X() + 3 * Observable.Y()])
    t2 = Observable.Sum([2 * Observable.X() + 3 * Observable.Y()])
    t3 = Observable.Sum([2 * Observable.Z() + 3 * Observable.H()])
    t4 = Observable.Sum([Observable.Z() + Observable.H()])
    assert t1 == t2
    assert t2 != t3
    assert t1 != t3
    assert t3 == t4


def test_invalid_target_length_for_sum_obs_term():
    observable, serialization_properties = (
        2 * Observable.Y() + 0.75 * Observable.Y() @ Observable.Z(),
        OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
    )
    target = [[0, 1], [0, 1]]

    invalid_target_len_for_term = "Invalid target for term 0 of Sum. Expected 1 targets, got 2"

    with pytest.raises(ValueError, match=invalid_target_len_for_term):
        observable.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )


def test_unscaled_tensor_product():
    observable = 3 * ((2 * Observable.X()) @ (5 * Observable.Y()))
    assert observable == 30 * (Observable.X() @ Observable.Y())
    assert observable._unscaled() == Observable.X() @ Observable.Y()



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from functools import reduce
from typing import Literal, Union

import braket.circuits.circuit as cir
from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.circuits.result_type import ResultType
from braket.circuits.text_diagram_builders.text_circuit_diagram import TextCircuitDiagram
from braket.registers.qubit_set import QubitSet


class AsciiCircuitDiagram(TextCircuitDiagram):
    """Builds ASCII string circuit diagrams."""

    @staticmethod
    def build_diagram(circuit: cir.Circuit) -> str:
        """Build a text circuit diagram.

        Args:
            circuit (Circuit): Circuit for which to build a diagram.

        Returns:
            str: string circuit diagram.
        """
        return AsciiCircuitDiagram._build(circuit)

    @classmethod
    def _vertical_delimiter(cls) -> str:
        """Character that connects qubits of multi-qubit gates."""
        return "|"

    @classmethod
    def _qubit_line_character(cls) -> str:
        """Character used for the qubit line."""
        return "-"

    @classmethod
    def _box_pad(cls) -> int:
        """number of blank space characters around the gate name."""
        return 0

    @classmethod
    def _qubit_line_spacing_above(cls) -> int:
        """number of empty lines above the qubit line."""
        return 1

    @classmethod
    def _qubit_line_spacing_below(cls) -> int:
        """number of empty lines below the qubit line."""
        return 0

    @classmethod
    def _duplicate_time_at_bottom(cls, lines: str) -> None:
        # duplicate times after an empty line
        lines.append(lines[0])

    @classmethod
    def _create_diagram_column(
        cls,
        circuit_qubits: QubitSet,
        items: list[Union[Instruction, ResultType]],
        global_phase: float | None = None,
    ) -> str:
        """Return a column in the ASCII string diagram of the circuit for a given list of items.

        Args:
            circuit_qubits (QubitSet): qubits in circuit
            items (list[Union[Instruction, ResultType]]): list of instructions or result types
            global_phase (float | None): the integrated global phase up to this column

        Returns:
            str: an ASCII string diagram for the specified moment in time for a column.
        """
        symbols = {qubit: cls._qubit_line_character() for qubit in circuit_qubits}
        connections = {qubit: "none" for qubit in circuit_qubits}

        for item in items:
            if isinstance(item, ResultType) and not item.target:
                target_qubits = circuit_qubits
                control_qubits = QubitSet()
                target_and_control = target_qubits.union(control_qubits)
                qubits = circuit_qubits
                ascii_symbols = [item.ascii_symbols[0]] * len(circuit_qubits)
            elif isinstance(item, Instruction) and isinstance(item.operator, CompilerDirective):
                target_qubits = circuit_qubits
                control_qubits = QubitSet()
                target_and_control = target_qubits.union(control_qubits)
                qubits = circuit_qubits
                ascii_symbol = item.ascii_symbols[0]
                marker = "*" * len(ascii_symbol)
                num_after = len(circuit_qubits) - 1
                after = ["|"] * (num_after - 1) + ([marker] if num_after else [])
                ascii_symbols = [ascii_symbol, *after]
            elif (
                isinstance(item, Instruction)
                and isinstance(item.operator, Gate)
                and item.operator.name == "GPhase"
            ):
                target_qubits = circuit_qubits
                control_qubits = QubitSet()
                target_and_control = QubitSet()
                qubits = circuit_qubits
                ascii_symbols = cls._qubit_line_character() * len(circuit_qubits)
            else:
                if isinstance(item.target, list):
                    target_qubits = reduce(QubitSet.union, map(QubitSet, item.target), QubitSet())
                else:
                    target_qubits = item.target
                control_qubits = getattr(item, "control", QubitSet())
                control_state = getattr(item, "control_state", "1" * len(control_qubits))
                map_control_qubit_states = {
                    qubit: state for qubit, state in zip(control_qubits, control_state)
                }

                target_and_control = target_qubits.union(control_qubits)
                qubits = QubitSet(range(min(target_and_control), max(target_and_control) + 1))

                ascii_symbols = item.ascii_symbols

            for qubit in qubits:
                # Determine if the qubit is part of the item or in the middle of a
                # multi qubit item.
                if qubit in target_qubits:
                    item_qubit_index = [
                        index for index, q in enumerate(target_qubits) if q == qubit
                    ][0]
                    power_string = (
                        f"^{power}"
                        if (
                            (power := getattr(item, "power", 1)) != 1
                            # this has the limitation of not printing the power
                            # when a user has a gate genuinely named C, but
                            # is necessary to enable proper printing of custom
                            # gates with built-in control qubits
                            and ascii_symbols[item_qubit_index] != "C"
                        )
                        else ""
                    )
                    symbols[qubit] = (
                        f"({ascii_symbols[item_qubit_index]}{power_string})"
                        if power_string
                        else ascii_symbols[item_qubit_index]
                    )
                elif qubit in control_qubits:
                    symbols[qubit] = "C" if map_control_qubit_states[qubit] else "N"
                else:
                    symbols[qubit] = "|"

                # Set the margin to be a connector if not on the first qubit
                if target_and_control and qubit != min(target_and_control):
                    connections[qubit] = "above"

        output = cls._create_output(symbols, connections, circuit_qubits, global_phase)
        return output

    # Ignore flake8 issue caused by Literal["above", "below", "both", "none"]
    # flake8: noqa: BCS005
    @classmethod
    def _draw_symbol(
        cls, symbol: str, symbols_width: int, connection: Literal["above", "below", "both", "none"]
    ) -> str:
        """Create a string representing the symbol.

        Args:
            symbol (str): the gate name
            symbols_width (int): size of the expected output. The ouput will be filled with
                cls._qubit_line_character() if needed.
            connection (Literal["above", "below", "both", "none"]): character indicating
                if the gate also involve a qubit with a lower index.

        Returns:
            str: a string representing the symbol.
        """
        connection_char = cls._vertical_delimiter() if connection in ["above"] else " "
        output = "{0:{width}}\n".format(connection_char, width=symbols_width + 1)
        output += "{0:{fill}{align}{width}}\n".format(
            symbol, fill=cls._qubit_line_character(), align="<", width=symbols_width + 1
        )
        return output



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from enum import Enum

import numpy as np

from braket.task_result import (
    AdditionalMetadata,
    AnalogHamiltonianSimulationTaskResult,
    TaskMetadata,
)


class AnalogHamiltonianSimulationShotStatus(str, Enum):
    SUCCESS = "Success"
    PARTIAL_SUCCESS = "Partial Success"
    FAILURE = "Failure"


@dataclass
class ShotResult:
    status: AnalogHamiltonianSimulationShotStatus
    pre_sequence: np.ndarray = None
    post_sequence: np.ndarray = None

    def __eq__(self, other: ShotResult) -> bool:
        if isinstance(other, ShotResult):
            return (
                self.status == other.status
                and _equal_sequences(self.pre_sequence, other.pre_sequence)
                and _equal_sequences(self.post_sequence, other.post_sequence)
            )
        return NotImplemented


@dataclass
class AnalogHamiltonianSimulationQuantumTaskResult:
    task_metadata: TaskMetadata
    additional_metadata: AdditionalMetadata
    measurements: list[ShotResult] = None

    def __eq__(self, other: AnalogHamiltonianSimulationQuantumTaskResult) -> bool:
        if isinstance(other, AnalogHamiltonianSimulationQuantumTaskResult):
            return (
                self.task_metadata.id == other.task_metadata.id
                and self.measurements == other.measurements
            )
        return NotImplemented

    @staticmethod
    def from_object(
        result: AnalogHamiltonianSimulationTaskResult,
    ) -> AnalogHamiltonianSimulationQuantumTaskResult:
        return AnalogHamiltonianSimulationQuantumTaskResult._from_object_internal(result)

    @staticmethod
    def from_string(result: str) -> AnalogHamiltonianSimulationQuantumTaskResult:
        return AnalogHamiltonianSimulationQuantumTaskResult._from_object_internal(
            AnalogHamiltonianSimulationTaskResult.parse_raw(result)
        )

    @classmethod
    def _from_object_internal(
        cls, result: AnalogHamiltonianSimulationTaskResult
    ) -> AnalogHamiltonianSimulationQuantumTaskResult:
        if result.measurements is not None:
            measurements = AnalogHamiltonianSimulationQuantumTaskResult._get_measurements(result)
        else:
            measurements = None
        return cls(
            task_metadata=result.taskMetadata,
            additional_metadata=result.additionalMetadata,
            measurements=measurements,
        )

    @classmethod
    def _get_measurements(cls, result: AnalogHamiltonianSimulationTaskResult) -> list[ShotResult]:
        measurements = []
        for measurement in result.measurements:
            status = AnalogHamiltonianSimulationShotStatus(measurement.shotMetadata.shotStatus)
            if measurement.shotResult.preSequence:
                pre_sequence = np.asarray(measurement.shotResult.preSequence, dtype=int)
            else:
                pre_sequence = None
            if measurement.shotResult.postSequence:
                post_sequence = np.asarray(measurement.shotResult.postSequence, dtype=int)
            else:
                post_sequence = None
            measurements.append(ShotResult(status, pre_sequence, post_sequence))
        return measurements

    def get_counts(self) -> dict[str, int]:
        """Aggregate state counts from AHS shot results.

        Notes:
            We use the following convention to denote the state of an atom (site).
            e: empty site
            r: Rydberg state atom
            g: ground state atom

        Returns:
            dict[str, int]: number of times each state configuration is measured.
            Returns None if none of shot measurements are successful.
            Only succesful shots contribute to the state count.
        """
        state_counts = Counter()
        states = ["e", "r", "g"]
        for shot in self.measurements:
            if shot.status == AnalogHamiltonianSimulationShotStatus.SUCCESS:
                pre = shot.pre_sequence
                post = shot.post_sequence
                # converting presequence and postsequence measurements to state_idx
                state_idx = [
                    0 if pre_i == 0 else 1 if post_i == 0 else 2 for pre_i, post_i in zip(pre, post)
                ]
                state = "".join(states[s_idx] for s_idx in state_idx)
                state_counts.update((state,))

        return dict(state_counts)

    def get_avg_density(self) -> np.ndarray:
        """Get the average Rydberg state densities from the result

        Returns:
            np.ndarray: The average densities from the result
        """
        counts = self.get_counts()

        N_ryd, N_ground = [], []
        for shot, count in counts.items():
            N_ryd.append([count if s == "r" else 0 for s in shot])
            N_ground.append([count if s == "g" else 0 for s in shot])

        N_ryd_cnt = np.sum(N_ryd, axis=0)
        N_ground_cnt = np.sum(N_ground, axis=0)

        avg_density = N_ryd_cnt / (N_ryd_cnt + N_ground_cnt)

        return avg_density


def _equal_sequences(sequence0: np.ndarray, sequence1: np.ndarray) -> bool:
    if sequence0 is None and sequence1 is None:
        return True
    if sequence0 is None or sequence1 is None:
        return False
    return np.allclose(sequence0, sequence1)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Observable, ResultType
from braket.circuits.noise_model import CriteriaKey, CriteriaKeyResult, ObservableCriteria


@pytest.mark.parametrize(
    "observables, qubits",
    [
        (None, range(3)),
        (None, None),
        (Observable.X, range(3)),
        ([Observable.X], range(3)),
        ([Observable.X, Observable.Y], range(3)),
        (Observable.X, 1),
        ([Observable.X], [1]),
        ([Observable.X, Observable.Y], [1, 0]),
        (Observable.X, None),
    ],
)
def test_happy_case(observables, qubits):
    criteria = ObservableCriteria(observables=observables, qubits=qubits)
    assert criteria.applicable_key_types() == [
        CriteriaKey.OBSERVABLE,
        CriteriaKey.QUBIT,
    ]
    if observables is None:
        assert CriteriaKeyResult.ALL == criteria.get_keys(CriteriaKey.OBSERVABLE)
    else:
        assert Observable.X in criteria.get_keys(CriteriaKey.OBSERVABLE)
    if qubits is None:
        assert CriteriaKeyResult.ALL == criteria.get_keys(CriteriaKey.QUBIT)
    else:
        assert 1 in criteria.get_keys(CriteriaKey.QUBIT)


@pytest.mark.parametrize(
    "observables, qubits",
    [
        (None, range(3)),
        (None, None),
        (Observable.X, range(3)),
        ([Observable.X], 1),
        ([Observable.X], [1]),
        ([Observable.X], [[1]]),
        ([Observable.X, Observable.Y], range(3)),
        ([Observable.X, Observable.Y], None),
    ],
)
def test_serialization(observables, qubits):
    test_criteria = ObservableCriteria(observables=observables, qubits=qubits)
    serialized_criteria = test_criteria.to_dict()
    assert serialized_criteria["__class__"] == "ObservableCriteria"
    assert "observables" in serialized_criteria
    assert "qubits" in serialized_criteria
    deserialized_criteria = ObservableCriteria.from_dict(serialized_criteria)
    assert test_criteria == deserialized_criteria


@pytest.mark.parametrize(
    "observables, qubits, matching_result_type, non_matching_result_type",
    [
        (
            None,
            range(3),
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.Z(), 2),
            ],
            [ResultType.Sample(Observable.X(), 4)],
        ),
        (
            None,
            None,
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.Z(), 10),
            ],
            [ResultType.Probability(0)],
        ),
        (
            Observable.X,
            range(3),
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.X(), 1),
                ResultType.Sample(Observable.X()),
            ],
            [
                ResultType.Sample(Observable.X(), 3),
                ResultType.Sample(Observable.Y(), 1),
            ],
        ),
        (
            Observable.X,
            1,
            [ResultType.Sample(Observable.X(), 1)],
            [ResultType.Sample(Observable.X(), 0)],
        ),
        (
            [Observable.X],
            [1],
            [ResultType.Sample(Observable.X(), 1)],
            [ResultType.Sample(Observable.X(), 0)],
        ),
        (
            [Observable.X, Observable.Y],
            range(3),
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.Y(), 2),
            ],
            [
                ResultType.Sample(Observable.X(), 3),
                ResultType.Sample(Observable.Z(), 1),
            ],
        ),
        (
            Observable.X,
            None,
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.X(), 10),
            ],
            [
                ResultType.Sample(Observable.Y(), 1),
                ResultType.Sample(Observable.Z(), 3),
            ],
        ),
        (
            Observable.X,
            [],
            [
                ResultType.Sample(Observable.X(), 0),
                ResultType.Sample(Observable.X(), 10),
            ],
            [
                ResultType.Sample(Observable.Y(), 1),
                ResultType.Sample(Observable.Z(), 3),
            ],
        ),
    ],
)
def test_matcher(observables, qubits, matching_result_type, non_matching_result_type):
    criteria = ObservableCriteria(observables=observables, qubits=qubits)
    for result_type in matching_result_type:
        assert criteria.result_type_matches(result_type)
    for instruction in non_matching_result_type:
        assert not criteria.result_type_matches(instruction)


@pytest.mark.parametrize(
    "observables, qubits",
    [
        ([Observable.X], [[0, 1]]),
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_invalid_params(observables, qubits):
    ObservableCriteria(observables=observables, qubits=qubits)


def test_representation():
    criteria = ObservableCriteria(observables=[Observable.X], qubits=0)
    str_representation = criteria.__repr__()
    assert str_representation == "ObservableCriteria(observables={'X'}, qubits={0})"


def test_string():
    criteria = ObservableCriteria(observables=[Observable.X], qubits=0)
    str_representation = criteria.__str__()
    assert str_representation == "ObservableCriteria({'X'}, {0})"


def test_get_keys_for_unknown_keytypes():
    criteria = ObservableCriteria(observables=[Observable.X], qubits=0)
    result = criteria.get_keys(CriteriaKey.UNITARY_GATE)
    assert len(result) == 0



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from datetime import timedelta
from decimal import Decimal
from functools import singledispatchmethod
from typing import Any

from braket.tracking.pricing import price_search
from braket.tracking.tracking_context import deregister_tracker, register_tracker
from braket.tracking.tracking_events import (
    _TaskCompletionEvent,
    _TaskCreationEvent,
    _TaskStatusEvent,
)

MIN_SIMULATOR_DURATION = timedelta(milliseconds=3000)


class Tracker:
    """Amazon Braket cost tracker.
    Use this class to track costs incurred from quantum tasks on Amazon Braket.
    """

    def __init__(self):
        self._resources = {}  # Key = quantum_task_arn

    def __enter__(self):
        register_tracker(self)
        return self

    def __exit__(self, *args):
        deregister_tracker(self)

    def start(self) -> Tracker:
        """Start tracking resources with this tracker.

        Returns:
            Tracker: self.
        """
        return self.__enter__()

    def stop(self) -> Tracker:
        """Stop tracking resources with this tracker.

        Returns:
            Tracker: self.
        """
        return self.__exit__()

    def receive_event(self, event: _TaskCreationEvent) -> None:
        """Process a Tack Creation Event.

        Args:
            event (_TaskCreationEvent): The event to process.
        """
        self._recieve_internal(event)

    def tracked_resources(self) -> list[str]:
        """Resources tracked by this tracker.

        Returns:
            list[str]: The list of quantum task ids for quantum tasks tracked by this tracker.
        """
        return list(self._resources.keys())

    def qpu_tasks_cost(self) -> Decimal:
        """Estimate cost of all quantum tasks tracked by this tracker that use Braket qpu devices.

        Note: Charges shown are estimates based on your Amazon Braket simulator and quantum
        processing unit (QPU) task usage. Estimated charges shown may differ from your actual
        charges. Estimated charges do not factor in any discounts or credits, and you may
        experience additional charges based on your use of other services such as
        Amazon Elastic Compute Cloud (Amazon EC2).

        Returns:
            Decimal: The estimated total cost in USD
        """
        total_cost = Decimal(0)
        for task_arn, details in self._resources.items():
            if "qpu" in details["device"]:
                total_cost = total_cost + _get_qpu_task_cost(task_arn, details)
        return total_cost

    def simulator_tasks_cost(self) -> Decimal:
        """Estimate cost of all quantum tasks tracked by this tracker using Braket simulator
         devices.

        Note: The cost of a simulator quantum task is not available until after the results for the
        task have been fetched. Call `result()` on an `AwsQuantumTask` before estimating its cost
        to ensure that the simulator usage is included in the cost estimate.

        Note: Charges shown are estimates based on your Amazon Braket simulator and quantum
        processing unit (QPU) task usage. Estimated charges shown may differ from your actual
        charges. Estimated charges do not factor in any discounts or credits, and you may
        experience additional charges based on your use of other services such as
        Amazon Elastic Compute Cloud (Amazon EC2).

        Returns:
            Decimal: The estimated total cost in USD
        """
        total_cost = Decimal(0)
        for task_arn, details in self._resources.items():
            if "simulator" in details["device"]:
                total_cost = total_cost + _get_simulator_task_cost(task_arn, details)
        return total_cost

    def quantum_tasks_statistics(self) -> dict[str, dict[str, Any]]:
        """Get a summary of quantum tasks grouped by device.

        Returns:
            dict[str, dict[str, Any]]: A dictionary where each key is a device arn, and maps to
            a dictionary sumarizing the quantum tasks run on the device. The summary includes the
            total shots sent to the device and the most recent status of the quantum tasks
            created on this device. For finished quantum tasks on simulator devices, the summary
            also includes the duration of the simulation.

        Example:
            >>> tracker.quantum_tasks_statistics()
            {'qpu_device_foo':
                {'shots' : 1000,
                 'tasks' : { 'COMPLETED' : 4,
                             'QUEUED' : 1 },
                },
             'simulator_device_bar':
                {'shots' : 1000
                 'tasks' : { 'COMPLETED' : 2,
                              'CREATED' : 1},
                 'execution_duration' : datetime.timedelta(seconds=5, microseconds=654321),
                 'billed_execution_duration' : datetime.timedelta(seconds=6, microseconds=123456)}}
        """
        stats = {}
        for _, details in self._resources.items():
            device_stats = stats.get(details["device"], {})

            shots = device_stats.get("shots", 0) + details["shots"]
            device_stats["shots"] = shots

            task_states = device_stats.get("tasks", {})
            task_states[details["status"]] = task_states.get(details["status"], 0) + 1
            device_stats["tasks"] = task_states

            if "execution_duration" in details:
                duration = (
                    device_stats.get("execution_duration", timedelta(0))
                    + details["execution_duration"]
                )
                billed_duration = (
                    timedelta(0)
                    if details.get("has_reservation_arn")
                    else (
                        device_stats.get("billed_execution_duration", timedelta(0))
                        + details["billed_duration"]
                    )
                )

                device_stats["execution_duration"] = duration
                device_stats["billed_execution_duration"] = billed_duration

            stats.setdefault(details["device"], {}).update(device_stats)

        return stats

    @singledispatchmethod
    def _recieve_internal(self, event: _TaskCreationEvent) -> None:
        raise ValueError(f"Event type {type(event)} is not supported")

    @_recieve_internal.register
    def _(self, event: _TaskCreationEvent) -> None:
        self._resources[event.arn] = {
            "shots": event.shots,
            "device": event.device,
            "status": "CREATED",
            "job_task": event.is_job_task,
        }

    @_recieve_internal.register
    def _(self, event: _TaskStatusEvent) -> None:
        resources = self._resources
        # Update task data corresponding to the arn only if it exists in resources
        if event.arn in resources:
            resources[event.arn]["status"] = event.status

    @_recieve_internal.register
    def _(self, event: _TaskCompletionEvent) -> None:
        resources = self._resources
        # Update task completion data corresponding to the arn only if it exists in resources
        if event.arn in resources:
            resources[event.arn]["status"] = event.status
            has_reservation_arn = event.has_reservation_arn
            resources[event.arn]["has_reservation_arn"] = has_reservation_arn
            if event.execution_duration:
                duration = timedelta(milliseconds=event.execution_duration)
                resources[event.arn]["execution_duration"] = duration
                resources[event.arn]["billed_duration"] = (
                    timedelta(milliseconds=0)
                    if has_reservation_arn
                    else max(duration, MIN_SIMULATOR_DURATION)
                )


def _get_qpu_task_cost(task_arn: str, details: dict) -> Decimal:
    if details["status"] in ["FAILED", "CANCELLED"] or details.get("has_reservation_arn"):
        return Decimal(0)
    task_region = task_arn.split(":")[3]

    search_dict = {"Region Code": task_region}

    device_name = details["device"].split("/")[-1]
    device_name = device_name[0].upper() + device_name[1:]
    if "2000Q" in device_name:
        device_name = "2000Q"
    elif "Advantage_system" in device_name:
        device_name = "Advantage_system"

    if details["job_task"]:
        search_dict["Device Name"] = device_name
        shot_product_family = "Braket Managed Jobs QPU Task Shot"
        task_product_family = "Braket Managed Jobs QPU Task"
    else:
        search_dict["DeviceName"] = device_name  # The difference in spelling is intentional
        shot_product_family = "Quantum Task-Shot"
        task_product_family = "Quantum Task"

    search_dict["Product Family"] = shot_product_family
    shot_prices = price_search(**search_dict)
    if len(shot_prices) != 1:
        raise ValueError(f"Found {len(shot_prices)} products matching {search_dict}")

    search_dict["Product Family"] = task_product_family
    task_prices = price_search(**search_dict)
    if len(task_prices) != 1:
        raise ValueError(f"Found {len(task_prices)} products matching {search_dict}")

    shot_price = shot_prices[0]
    task_price = task_prices[0]

    for price in [shot_price, task_price]:
        if price["Currency"] != "USD":
            raise ValueError(f"Expected USD, found {price['Currency']}")

    shot_cost = Decimal(shot_price["PricePerUnit"]) * details["shots"]
    task_cost = Decimal(task_price["PricePerUnit"]) * 1

    return shot_cost + task_cost


def _get_simulator_task_cost(task_arn: str, details: dict) -> Decimal:
    if not details.get("billed_duration"):
        return Decimal(0)
    task_region = task_arn.split(":")[3]

    device_name = details["device"].split("/")[-1].upper()

    if details["job_task"]:
        product_family = "Braket Managed Jobs Simulator Task"
        operation = "Managed-Jobs"
    else:
        product_family = "Simulator Task"
        operation = "CompleteTask"
        if details["status"] == "FAILED" and device_name == "TN1":
            # Rehersal step of TN1 can fail and charges still apply.
            operation = "FailedTask"

    search_dict = {
        "Region Code": task_region,
        "Version": device_name,
        "Product Family": product_family,
        "operation": operation,
    }

    duration_prices = price_search(**search_dict)

    if len(duration_prices) != 1:
        raise ValueError(f"Found {len(duration_prices)} products matching {search_dict}")

    duration_price = duration_prices[0]

    if duration_price["Currency"] != "USD":
        raise ValueError(f"Expected USD, found {duration_price['Currency']}")

    duration_cost = (
        Decimal(duration_price["PricePerUnit"])
        * Decimal(details["billed_duration"] / timedelta(milliseconds=1))
        / Decimal(timedelta(**{duration_price["Unit"]: 1}) / timedelta(milliseconds=1))
    )

    return duration_cost



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import re
from functools import reduce
from typing import Union

import braket.ir.jaqcd as ir
from braket.circuits import circuit
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.observable import Observable
from braket.circuits.observables import Sum
from braket.circuits.result_type import (
    ObservableParameterResultType,
    ObservableResultType,
    ResultType,
)
from braket.circuits.serialization import IRType, OpenQASMSerializationProperties
from braket.registers.qubit_set import QubitSet, QubitSetInput

"""
To add a new result type:
    1. Implement the class and extend `ResultType`
    2. Add a method with the `@circuit.subroutine(register=True)` decorator. Method name
       is added into the `Circuit` class. This method is the default way
       clients add this result type to a circuit.
    3. Register the class with the `ResultType` class via `ResultType.register_result_type()`.
"""


class StateVector(ResultType):
    """The full state vector as a requested result type.
    This is available on simulators only when `shots=0`.
    """

    def __init__(self):
        super().__init__(ascii_symbols=["StateVector"])

    def _to_jaqcd(self) -> ir.StateVector:
        return ir.StateVector.construct()

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        return "#pragma braket result state_vector"

    @staticmethod
    @circuit.subroutine(register=True)
    def state_vector() -> ResultType:
        """Registers this function into the circuit class.

        Returns:
            ResultType: state vector as a requested result type

        Examples:
            >>> circ = Circuit().state_vector()
        """
        return ResultType.StateVector()

    def __eq__(self, other: StateVector) -> bool:
        if isinstance(other, StateVector):
            return True
        return False

    def __copy__(self) -> StateVector:
        return type(self)()

    # must redefine __hash__ since __eq__ is overwritten
    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
    def __hash__(self) -> int:
        return super().__hash__()


ResultType.register_result_type(StateVector)


class DensityMatrix(ResultType):
    """The full density matrix as a requested result type.
    This is available on simulators only when `shots=0`.
    """

    def __init__(self, target: QubitSetInput | None = None):
        """Inits a `DensityMatrix`.

        Args:
            target (QubitSetInput | None): The target qubits
                of the reduced density matrix. Default is `None`, and the
                full density matrix is returned.

        Examples:
            >>> ResultType.DensityMatrix(target=[0, 1])
        """
        self._target = QubitSet(target)
        ascii_symbols = ["DensityMatrix"] * len(self._target) if self._target else ["DensityMatrix"]
        super().__init__(ascii_symbols=ascii_symbols)

    @property
    def target(self) -> QubitSet:
        return self._target

    @target.setter
    def target(self, target: QubitSetInput) -> None:
        """Sets the target qubit set.

        Args:
            target (QubitSetInput): The target qubit set.
        """
        self._target = QubitSet(target)

    def _to_jaqcd(self) -> ir.DensityMatrix:
        if self.target:
            # convert qubits to int as required by the ir type
            return ir.DensityMatrix.construct(targets=[int(qubit) for qubit in self.target])
        else:
            return ir.DensityMatrix.construct()

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        if not self.target:
            return "#pragma braket result density_matrix all"
        targets = ", ".join(
            serialization_properties.format_target(int(target)) for target in self.target
        )
        return f"#pragma braket result density_matrix {targets}"

    @staticmethod
    @circuit.subroutine(register=True)
    def density_matrix(target: QubitSetInput | None = None) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput | None): The target qubits
                of the reduced density matrix. Default is `None`, and the
                full density matrix is returned.

        Returns:
            ResultType: density matrix as a requested result type

        Examples:
            >>> circ = Circuit().density_matrix(target=[0, 1])
        """
        return ResultType.DensityMatrix(target=target)

    def __eq__(self, other: DensityMatrix) -> bool:
        if isinstance(other, DensityMatrix):
            return self.target == other.target
        return False

    def __repr__(self) -> str:
        return f"DensityMatrix(target={self.target})"

    def __copy__(self) -> DensityMatrix:
        return type(self)(target=self.target)

    # must redefine __hash__ since __eq__ is overwritten
    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
    def __hash__(self) -> int:
        return super().__hash__()


ResultType.register_result_type(DensityMatrix)


class AdjointGradient(ObservableParameterResultType):
    """The gradient of the expectation value of the provided observable, applied to target,
    with respect to the given parameter.
    """

    def __init__(
        self,
        observable: Observable,
        target: list[QubitSetInput] | None = None,
        parameters: list[Union[str, FreeParameter]] | None = None,
    ):
        """Inits an `AdjointGradient`.

        Args:
            observable (Observable): The expectation value of this observable is the function
                against which parameters in the gradient are differentiated.
            target (list[QubitSetInput] | None): Target qubits that the result type is requested
                for. Each term in the target list should have the same number of qubits as the
                corresponding term in the observable. Default is `None`, which means the
                observable must operate only on 1 qubit and it is applied to all qubits
                in parallel.
            parameters (list[Union[str, FreeParameter]] | None): The free parameters in the circuit
                to differentiate with respect to. Default: `all`.

        Raises:
            ValueError: If the observable's qubit count does not equal the number of target
                qubits, or if `target=None` and the observable's qubit count is not 1.


        Examples:
            >>> ResultType.AdjointGradient(observable=Observable.Z(),
                                        target=0, parameters=["alpha", "beta"])

            >>> tensor_product = Observable.Y() @ Observable.Z()
            >>> hamiltonian = Observable.Y() @ Observable.Z() + Observable.H()
            >>> ResultType.AdjointGradient(
            >>>     observable=tensor_product,
            >>>     target=[[0, 1], [2]],
            >>>     parameters=["alpha", "beta"],
            >>> )
        """
        if isinstance(observable, Sum):
            target_qubits = reduce(QubitSet.union, map(QubitSet, target), QubitSet())
        else:
            target_qubits = QubitSet(target)

        super().__init__(
            ascii_symbols=[f"AdjointGradient({observable.ascii_symbols[0]})"] * len(target_qubits),
            observable=observable,
            target=target,
            parameters=parameters,
        )

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        observable_ir = self.observable.to_ir(
            target=self.target,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )

        pragma_parameters = ", ".join(self.parameters) if self.parameters else "all"

        return (
            f"#pragma braket result adjoint_gradient "
            f"expectation({observable_ir}) {pragma_parameters}"
        )

    @staticmethod
    @circuit.subroutine(register=True)
    def adjoint_gradient(
        observable: Observable,
        target: list[QubitSetInput] | None = None,
        parameters: list[Union[str, FreeParameter]] | None = None,
    ) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            observable (Observable): The expectation value of this observable is the function
                against which parameters in the gradient are differentiated.
            target (list[QubitSetInput] | None): Target qubits that the result type is requested
                for. Each term in the target list should have the same number of qubits as the
                corresponding term in the observable. Default is `None`, which means the
                observable must operate only on 1 qubit and it is applied to all qubits
                in parallel.
            parameters (list[Union[str, FreeParameter]] | None): The free parameters in the circuit
                to differentiate with respect to. Default: `all`.

        Returns:
            ResultType: gradient computed via adjoint differentiation as a requested result type

        Examples:
            >>> alpha, beta = FreeParameter('alpha'), FreeParameter('beta')
            >>> circ = Circuit().h(0).h(1).rx(0, alpha).yy(0, 1, beta).adjoint_gradient(
            >>>     observable=Observable.Z(), target=[0], parameters=[alpha, beta]
            >>> )
        """
        return ResultType.AdjointGradient(
            observable=observable, target=target, parameters=parameters
        )


ResultType.register_result_type(AdjointGradient)


class Amplitude(ResultType):
    """The amplitude of the specified quantum states as a requested result type.
    This is available on simulators only when `shots=0`.
    """

    def __init__(self, state: list[str]):
        """Initializes an `Amplitude`.

        Args:
            state (list[str]): list of quantum states as strings with "0" and "1"

        Raises:
            ValueError: If state is `None` or an empty list, or
                state is not a list of strings of '0' and '1'

        Examples:
            >>> ResultType.Amplitude(state=['01', '10'])
        """
        if (
            not state
            or not isinstance(state, list)
            or not all(
                isinstance(amplitude, str) and re.fullmatch("^[01]+$", amplitude)
                for amplitude in state
            )
        ):
            raise ValueError(
                "A non-empty list of states must be specified in binary encoding e.g. ['01', '10']"
            )
        super().__init__(ascii_symbols=[f"Amplitude({','.join(state)})"])
        self._state = state

    @property
    def state(self) -> list[str]:
        return self._state

    def _to_jaqcd(self) -> ir.Amplitude:
        return ir.Amplitude.construct(states=self.state)

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        states = ", ".join(f'"{state}"' for state in self.state)
        return f"#pragma braket result amplitude {states}"

    @staticmethod
    @circuit.subroutine(register=True)
    def amplitude(state: list[str]) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            state (list[str]): list of quantum states as strings with "0" and "1"

        Returns:
            ResultType: state vector as a requested result type

        Examples:
            >>> circ = Circuit().amplitude(state=["01", "10"])
        """
        return ResultType.Amplitude(state=state)

    def __eq__(self, other: Amplitude):
        if isinstance(other, Amplitude):
            return self.state == other.state
        return False

    def __repr__(self):
        return f"Amplitude(state={self.state})"

    def __copy__(self):
        return type(self)(state=self.state)

    def __hash__(self) -> int:
        return super().__hash__()


ResultType.register_result_type(Amplitude)


class Probability(ResultType):
    """Probability in the computational basis as the requested result type.

    It can be the probability of all states if no targets are specified, or the marginal
    probability of a restricted set of states if only a subset of all qubits are specified as
    targets.

    For `shots>0`, this is calculated by measurements. For `shots=0`, this is supported
    only on simulators and represents the exact result.
    """

    def __init__(self, target: QubitSetInput | None = None):
        """Inits a `Probability`.

        Args:
            target (QubitSetInput | None): The target qubits that the
                result type is requested for. Default is `None`, which means all qubits for the
                circuit.

        Examples:
            >>> ResultType.Probability(target=[0, 1])
        """
        self._target = QubitSet(target)
        ascii_symbols = ["Probability"] * len(self._target) if self._target else ["Probability"]
        super().__init__(ascii_symbols=ascii_symbols)

    @property
    def target(self) -> QubitSet:
        return self._target

    @target.setter
    def target(self, target: QubitSetInput) -> None:
        """Sets the target qubit set.

        Args:
            target (QubitSetInput): The target qubit set.
        """
        self._target = QubitSet(target)

    def _to_jaqcd(self) -> ir.Probability:
        if self.target:
            # convert qubits to int as required by the ir type
            return ir.Probability.construct(targets=[int(qubit) for qubit in self.target])
        else:
            return ir.Probability.construct()

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        if not self.target:
            return "#pragma braket result probability all"
        targets = ", ".join(
            serialization_properties.format_target(int(target)) for target in self.target
        )
        return f"#pragma braket result probability {targets}"

    @staticmethod
    @circuit.subroutine(register=True)
    def probability(target: QubitSetInput | None = None) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput | None): The target qubits that the
                result type is requested for. Default is `None`, which means all qubits for the
                circuit.

        Returns:
            ResultType: probability as a requested result type

        Examples:
            >>> circ = Circuit().probability(target=[0, 1])
        """
        return ResultType.Probability(target=target)

    def __eq__(self, other: Probability) -> bool:
        if isinstance(other, Probability):
            return self.target == other.target
        return False

    def __repr__(self) -> str:
        return f"Probability(target={self.target})"

    def __copy__(self) -> Probability:
        return type(self)(target=self.target)

    def __hash__(self) -> int:
        return super().__hash__()


ResultType.register_result_type(Probability)


class Expectation(ObservableResultType):
    """Expectation of the specified target qubit set and observable as the requested result type.

    If no targets are specified, the observable must operate only on 1 qubit and it
    is applied to all qubits in parallel. Otherwise, the number of specified targets
    must be equivalent to the number of qubits the observable can be applied to.

    For `shots>0`, this is calculated by measurements. For `shots=0`, this is supported
    only by simulators and represents the exact result.

    See :mod:`braket.circuits.observables` module for all of the supported observables.
    """

    def __init__(self, observable: Observable, target: QubitSetInput | None = None):
        """Inits an `Expectation`.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                operate only on 1 qubit and it is applied to all qubits in parallel.


        Examples:
            >>> ResultType.Expectation(observable=Observable.Z(), target=0)

            >>> tensor_product = Observable.Y() @ Observable.Z()
            >>> ResultType.Expectation(observable=tensor_product, target=[0, 1])
        """
        super().__init__(
            ascii_symbols=[f"Expectation({obs_ascii})" for obs_ascii in observable.ascii_symbols],
            observable=observable,
            target=target,
        )

    def _to_jaqcd(self) -> ir.Expectation:
        if self.target:
            return ir.Expectation.construct(
                observable=self.observable.to_ir(), targets=[int(qubit) for qubit in self.target]
            )
        else:
            return ir.Expectation.construct(observable=self.observable.to_ir())

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        observable_ir = self.observable.to_ir(
            target=self.target,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        return f"#pragma braket result expectation {observable_ir}"

    @staticmethod
    @circuit.subroutine(register=True)
    def expectation(observable: Observable, target: QubitSetInput | None = None) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                operate only on 1 qubit and it is applied to all qubits in parallel.

        Returns:
            ResultType: expectation as a requested result type

        Examples:
            >>> circ = Circuit().expectation(observable=Observable.Z(), target=0)
        """
        return ResultType.Expectation(observable=observable, target=target)


ResultType.register_result_type(Expectation)


class Sample(ObservableResultType):
    """Sample of specified target qubit set and observable as the requested result type.

    If no targets are specified, the observable must operate only on 1 qubit and it
    is applied to all qubits in parallel. Otherwise, the number of specified targets
    must equal the number of qubits the observable can be applied to.

    This is only available for `shots>0`.

    See :mod:`braket.circuits.observables` module for all of the supported observables.
    """

    def __init__(self, observable: Observable, target: QubitSetInput | None = None):
        """Inits a `Sample`.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                operate only on 1 qubit and it is applied to all qubits in parallel.

        Examples:
            >>> ResultType.Sample(observable=Observable.Z(), target=0)

            >>> tensor_product = Observable.Y() @ Observable.Z()
            >>> ResultType.Sample(observable=tensor_product, target=[0, 1])
        """
        super().__init__(
            ascii_symbols=[f"Sample({obs_ascii})" for obs_ascii in observable.ascii_symbols],
            observable=observable,
            target=target,
        )

    def _to_jaqcd(self) -> ir.Sample:
        if self.target:
            return ir.Sample.construct(
                observable=self.observable.to_ir(), targets=[int(qubit) for qubit in self.target]
            )
        else:
            return ir.Sample.construct(observable=self.observable.to_ir())

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        observable_ir = self.observable.to_ir(
            target=self.target,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        return f"#pragma braket result sample {observable_ir}"

    @staticmethod
    @circuit.subroutine(register=True)
    def sample(observable: Observable, target: QubitSetInput | None = None) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                operate only on 1 qubit and it is applied to all qubits in parallel.

        Returns:
            ResultType: sample as a requested result type

        Examples:
            >>> circ = Circuit().sample(observable=Observable.Z(), target=0)
        """
        return ResultType.Sample(observable=observable, target=target)


ResultType.register_result_type(Sample)


class Variance(ObservableResultType):
    """Variance of specified target qubit set and observable as the requested result type.

    If no targets are specified, the observable must operate only on 1 qubit and it
    is applied to all qubits in parallel. Otherwise, the number of targets specified
    must equal the number of qubits that the observable can be applied to.

    For `shots>0`, this is calculated by measurements. For `shots=0`, this is supported
    only by simulators and represents the exact result.

    See :mod:`braket.circuits.observables` module for all of the supported observables.
    """

    def __init__(self, observable: Observable, target: QubitSetInput | None = None):
        """Inits a `Variance`.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                operate only on 1 qubit and it is applied to all qubits in parallel.

        Raises:
            ValueError: If the observable's qubit count does not equal the number of target
                qubits, or if `target=None` and the observable's qubit count is not 1.

        Examples:
            >>> ResultType.Variance(observable=Observable.Z(), target=0)

            >>> tensor_product = Observable.Y() @ Observable.Z()
            >>> ResultType.Variance(observable=tensor_product, target=[0, 1])
        """
        super().__init__(
            ascii_symbols=[f"Variance({obs_ascii})" for obs_ascii in observable.ascii_symbols],
            observable=observable,
            target=target,
        )

    def _to_jaqcd(self) -> ir.Variance:
        if self.target:
            return ir.Variance.construct(
                observable=self.observable.to_ir(), targets=[int(qubit) for qubit in self.target]
            )
        else:
            return ir.Variance.construct(observable=self.observable.to_ir())

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        observable_ir = self.observable.to_ir(
            target=self.target,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        return f"#pragma braket result variance {observable_ir}"

    @staticmethod
    @circuit.subroutine(register=True)
    def variance(observable: Observable, target: QubitSetInput | None = None) -> ResultType:
        """Registers this function into the circuit class.

        Args:
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                only operate on 1 qubit and it will be applied to all qubits in parallel

        Returns:
            ResultType: variance as a requested result type

        Examples:
            >>> circ = Circuit().variance(observable=Observable.Z(), target=0)
        """
        return ResultType.Variance(observable=observable, target=target)


ResultType.register_result_type(Variance)



import pytest

from braket.circuits.basis_state import BasisState


@pytest.mark.parametrize(
    "basis_state_input, size, as_tuple, as_int, as_string",
    (
        (
            [1, 0, 1],
            None,
            (1, 0, 1),
            5,
            "101",
        ),
        (
            [1, 0, 1],
            5,
            (0, 0, 1, 0, 1),
            5,
            "00101",
        ),
        (
            "1",
            3,
            (0, 0, 1),
            1,
            "001",
        ),
        (
            "101",
            None,
            (1, 0, 1),
            5,
            "101",
        ),
        (
            5,
            None,
            (1, 0, 1),
            5,
            "101",
        ),
        (
            5,
            4,
            (0, 1, 0, 1),
            5,
            "0101",
        ),
    ),
)
def test_as_props(basis_state_input, size, as_tuple, as_int, as_string):
    basis_state = BasisState(basis_state_input, size)
    assert basis_state.as_tuple == as_tuple
    assert basis_state.as_int == as_int
    assert basis_state.as_string == as_string == str(basis_state)
    assert repr(basis_state) == f'BasisState("{as_string}")'


@pytest.mark.parametrize(
    "basis_state_input, index, substate_input",
    (
        (
            "1001",
            slice(None),
            "1001",
        ),
        (
            "1001",
            3,
            "1",
        ),
        (
            "1010",
            slice(None, None, 2),
            "11",
        ),
        (
            "1010",
            slice(1, None, 2),
            "00",
        ),
        (
            "1010",
            slice(None, -2),
            "10",
        ),
        (
            "1010",
            -1,
            "0",
        ),
    ),
)
def test_indexing(basis_state_input, index, substate_input):
    assert BasisState(basis_state_input)[index] == BasisState(substate_input)


def test_bool():
    assert all(
        [
            BasisState("100"),
            BasisState("111"),
            BasisState("1"),
        ]
    )
    assert not BasisState("0")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import os


def get_job_name() -> str:
    """Get the name of the current job.

    Returns:
        str: The name of the job if in a job, else an empty string.
    """
    return os.getenv("AMZN_BRAKET_JOB_NAME", "")


def get_job_device_arn() -> str:
    """Get the device ARN of the current job. If not in a job, default to "local:none/none".

    Returns:
        str: The device ARN of the current job or "local:none/none".
    """
    return os.getenv("AMZN_BRAKET_DEVICE_ARN", "local:none/none")


def get_input_data_dir(channel: str = "input") -> str:
    """Get the job input data directory.

    Args:
        channel (str): The name of the input channel. Default value
            corresponds to the default input channel name, `input`.

    Returns:
        str: The input directory, defaulting to current working directory.
    """
    input_dir = os.getenv("AMZN_BRAKET_INPUT_DIR", ".")
    if input_dir != ".":
        return f"{input_dir}/{channel}"
    return input_dir


def get_results_dir() -> str:
    """Get the job result directory.

    Returns:
        str: The results directory, defaulting to current working directory.
    """
    return os.getenv("AMZN_BRAKET_JOB_RESULTS_DIR", ".")


def get_checkpoint_dir() -> str:
    """Get the job checkpoint directory.

    Returns:
        str: The checkpoint directory, defaulting to current working directory.
    """
    return os.getenv("AMZN_BRAKET_CHECKPOINT_DIR", ".")


def get_hyperparameters() -> dict[str, str]:
    """Get the job hyperparameters as a dict, with the values stringified.

    Returns:
        dict[str, str]: The hyperparameters of the job.
    """
    if "AMZN_BRAKET_HP_FILE" in os.environ:
        with open(os.getenv("AMZN_BRAKET_HP_FILE")) as f:
            return json.load(f)
    return {}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import codecs
import pickle
from typing import Any

from braket.jobs_data import PersistedJobDataFormat


def serialize_values(
    data_dictionary: dict[str, Any], data_format: PersistedJobDataFormat
) -> dict[str, Any]:
    """Serializes the `data_dictionary` values to the format specified by `data_format`.

    Args:
        data_dictionary (dict[str, Any]): Dict whose values are to be serialized.
        data_format (PersistedJobDataFormat): The data format used to serialize the
            values. Note that for `PICKLED` data formats, the values are base64 encoded
            after serialization, so that they represent valid UTF-8 text and are compatible
            with `PersistedJobData.json()`.

    Returns:
        dict[str, Any]: Dict with same keys as `data_dictionary` and values serialized to
        the specified `data_format`.
    """
    return (
        {
            k: codecs.encode(pickle.dumps(v, protocol=4), "base64").decode()
            for k, v in data_dictionary.items()
        }
        if data_format == PersistedJobDataFormat.PICKLED_V4
        else data_dictionary
    )


def deserialize_values(
    data_dictionary: dict[str, Any], data_format: PersistedJobDataFormat
) -> dict[str, Any]:
    """Deserializes the `data_dictionary` values from the format specified by `data_format`.

    Args:
        data_dictionary (dict[str, Any]): Dict whose values are to be deserialized.
        data_format (PersistedJobDataFormat): The data format that the `data_dictionary` values
            are currently serialized with.

    Returns:
        dict[str, Any]: Dict with same keys as `data_dictionary` and values deserialized from
        the specified `data_format` to plaintext.
    """
    return (
        {k: pickle.loads(codecs.decode(v.encode(), "base64")) for k, v in data_dictionary.items()}
        if data_format == PersistedJobDataFormat.PICKLED_V4
        else data_dictionary
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import importlib
import json
import os
import urllib.request
import warnings
from datetime import datetime
from enum import Enum
from typing import Any, ClassVar, Optional, Union

from botocore.errorfactory import ClientError
from networkx import DiGraph, complete_graph, from_edgelist

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.aws.aws_quantum_task import AwsQuantumTask
from braket.aws.aws_quantum_task_batch import AwsQuantumTaskBatch
from braket.aws.aws_session import AwsSession
from braket.aws.queue_information import QueueDepthInfo, QueueType
from braket.circuits import Circuit, Gate, QubitSet
from braket.circuits.gate_calibrations import GateCalibrations
from braket.circuits.noise_model import NoiseModel
from braket.device_schema import DeviceCapabilities, ExecutionDay, GateModelQpuParadigmProperties
from braket.device_schema.dwave import DwaveProviderProperties

# TODO: Remove device_action module once this is added to init in the schemas repo
from braket.device_schema.pulse.pulse_device_action_properties_v1 import PulseDeviceActionProperties
from braket.devices.device import Device
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQasmProgram
from braket.parametric.free_parameter import FreeParameter
from braket.parametric.free_parameter_expression import _is_float
from braket.pulse import ArbitraryWaveform, Frame, Port, PulseSequence
from braket.pulse.waveforms import _parse_waveform_from_calibration_schema
from braket.schema_common import BraketSchemaBase


class AwsDeviceType(str, Enum):
    """Possible AWS device types"""

    SIMULATOR = "SIMULATOR"
    QPU = "QPU"


class AwsDevice(Device):
    """Amazon Braket implementation of a device.
    Use this class to retrieve the latest metadata about the device and to run a quantum task on the
    device.
    """

    REGIONS = ("us-east-1", "us-west-1", "us-west-2", "eu-west-2")

    DEFAULT_SHOTS_QPU = 1000
    DEFAULT_SHOTS_SIMULATOR = 0
    DEFAULT_MAX_PARALLEL = 10

    _GET_DEVICES_ORDER_BY_KEYS = frozenset({"arn", "name", "type", "provider_name", "status"})

    _RIGETTI_GATES_TO_BRAKET: ClassVar[dict[str, str | None]] = {
        # Rx_12 does not exist in the Braket SDK, it is a gate between |1> and |2>.
        "Rx_12": None,
        "Cz": "CZ",
        "Cphaseshift": "CPhaseShift",
        "Xy": "XY",
    }

    def __init__(
        self,
        arn: str,
        aws_session: Optional[AwsSession] = None,
        noise_model: Optional[NoiseModel] = None,
    ):
        """Initializes an `AwsDevice`.

        Args:
            arn (str): The ARN of the device
            aws_session (Optional[AwsSession]): An AWS session object. Default is `None`.
            noise_model (Optional[NoiseModel]): The Braket noise model to apply to the circuit
                before execution. Noise model can only be added to the devices that support
                noise simulation.

        Note:
            Some devices (QPUs) are physically located in specific AWS Regions. In some cases,
            the current `aws_session` connects to a Region other than the Region in which the QPU is
            physically located. When this occurs, a cloned `aws_session` is created for the Region
            the QPU is located in.

            See `braket.aws.aws_device.AwsDevice.REGIONS` for the AWS regions provider
            devices are located in across the AWS Braket service.
            This is not a device specific tuple.
        """
        super().__init__(name=None, status=None)
        self._arn = arn
        self._gate_calibrations = None
        self._properties = None
        self._provider_name = None
        self._poll_interval_seconds = None
        self._type = None
        self._aws_session = self._get_session_and_initialize(aws_session or AwsSession())
        self._ports = None
        self._frames = None
        if noise_model:
            self._validate_device_noise_model_support(noise_model)
        self._noise_model = noise_model

    def run(
        self,
        task_specification: Union[
            Circuit,
            Problem,
            OpenQasmProgram,
            BlackbirdProgram,
            PulseSequence,
            AnalogHamiltonianSimulation,
        ],
        s3_destination_folder: Optional[AwsSession.S3DestinationFolder] = None,
        shots: Optional[int] = None,
        poll_timeout_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: Optional[float] = None,
        inputs: Optional[dict[str, float]] = None,
        gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]] = None,
        reservation_arn: str | None = None,
        *aws_quantum_task_args: Any,
        **aws_quantum_task_kwargs: Any,
    ) -> AwsQuantumTask:
        """Run a quantum task specification on this device. A quantum task can be a circuit or an
        annealing problem.

        Args:
            task_specification (Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, PulseSequence, AnalogHamiltonianSimulation]):
                Specification of quantum task (circuit, OpenQASM program or AHS program)
                to run on device.
            s3_destination_folder (Optional[S3DestinationFolder]): The S3 location to
                save the quantum task's results to. Default is `<default_bucket>/tasks` if evoked outside a
                Braket Hybrid Job, `<Job Bucket>/jobs/<job name>/tasks` if evoked inside a Braket Hybrid Job.
            shots (Optional[int]): The number of times to run the circuit or annealing problem.
                Default is 1000 for QPUs and 0 for simulators.
            poll_timeout_seconds (float): The polling timeout for `AwsQuantumTask.result()`,
                in seconds. Default: 5 days.
            poll_interval_seconds (Optional[float]): The polling interval for `AwsQuantumTask.result()`,
                in seconds. Defaults to the ``getTaskPollIntervalMillis`` value specified in
                ``self.properties.service`` (divided by 1000) if provided, otherwise 1 second.
            inputs (Optional[dict[str, float]]): Inputs to be passed along with the
                IR. If the IR supports inputs, the inputs will be updated with this value.
                Default: {}.
            gate_definitions (Optional[dict[tuple[Gate, QubitSet], PulseSequence]]): A
                `dict[tuple[Gate, QubitSet], PulseSequence]]` for a user defined gate calibration.
                The calibration is defined for a particular `Gate` on a particular `QubitSet`
                and is represented by a `PulseSequence`.
                Default: None.
            reservation_arn (str | None): The reservation ARN provided by Braket Direct
                to reserve exclusive usage for the device to run the quantum task on.
                Note: If you are creating tasks in a job that itself was created reservation ARN,
                those tasks do not need to be created with the reservation ARN.
                Default: None.
            *aws_quantum_task_args (Any): Arbitrary arguments.
            **aws_quantum_task_kwargs (Any): Arbitrary keyword arguments.

        Returns:
            AwsQuantumTask: An AwsQuantumTask that tracks the execution on the device.

        Examples:
            >>> circuit = Circuit().h(0).cnot(0, 1)
            >>> device = AwsDevice("arn1")
            >>> device.run(circuit, ("bucket-foo", "key-bar"))

            >>> circuit = Circuit().h(0).cnot(0, 1)
            >>> device = AwsDevice("arn2")
            >>> device.run(task_specification=circuit,
            >>>     s3_destination_folder=("bucket-foo", "key-bar"))

            >>> circuit = Circuit().h(0).cnot(0, 1)
            >>> device = AwsDevice("arn3")
            >>> device.run(task_specification=circuit,
            >>>     s3_destination_folder=("bucket-foo", "key-bar"), disable_qubit_rewiring=True)

            >>> problem = Problem(
            >>>     ProblemType.ISING,
            >>>     linear={1: 3.14},
            >>>     quadratic={(1, 2): 10.08},
            >>> )
            >>> device = AwsDevice("arn4")
            >>> device.run(problem, ("bucket-foo", "key-bar"),
            >>>     device_parameters={
            >>>         "providerLevelParameters": {"postprocessingType": "SAMPLING"}}
            >>> )

        See Also:
            `braket.aws.aws_quantum_task.AwsQuantumTask.create()`
        """  # noqa E501
        if self._noise_model:
            task_specification = self._apply_noise_model_to_circuit(task_specification)
        return AwsQuantumTask.create(
            self._aws_session,
            self._arn,
            task_specification,
            s3_destination_folder
            or (
                AwsSession.parse_s3_uri(os.environ.get("AMZN_BRAKET_TASK_RESULTS_S3_URI"))
                if "AMZN_BRAKET_TASK_RESULTS_S3_URI" in os.environ
                else None
            )
            or (self._aws_session.default_bucket(), "tasks"),
            shots if shots is not None else self._default_shots,
            poll_timeout_seconds=poll_timeout_seconds,
            poll_interval_seconds=poll_interval_seconds or self._poll_interval_seconds,
            inputs=inputs,
            gate_definitions=gate_definitions,
            reservation_arn=reservation_arn,
            *aws_quantum_task_args,
            **aws_quantum_task_kwargs,
        )

    def run_batch(
        self,
        task_specifications: Union[
            Union[
                Circuit,
                Problem,
                OpenQasmProgram,
                BlackbirdProgram,
                PulseSequence,
                AnalogHamiltonianSimulation,
            ],
            list[
                Union[
                    Circuit,
                    Problem,
                    OpenQasmProgram,
                    BlackbirdProgram,
                    PulseSequence,
                    AnalogHamiltonianSimulation,
                ]
            ],
        ],
        s3_destination_folder: Optional[AwsSession.S3DestinationFolder] = None,
        shots: Optional[int] = None,
        max_parallel: Optional[int] = None,
        max_connections: int = AwsQuantumTaskBatch.MAX_CONNECTIONS_DEFAULT,
        poll_timeout_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        inputs: Optional[Union[dict[str, float], list[dict[str, float]]]] = None,
        gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]] = None,
        reservation_arn: Optional[str] = None,
        *aws_quantum_task_args,
        **aws_quantum_task_kwargs,
    ) -> AwsQuantumTaskBatch:
        """Executes a batch of quantum tasks in parallel

        Args:
            task_specifications (Union[Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, PulseSequence, AnalogHamiltonianSimulation], list[Union[ Circuit, Problem, OpenQasmProgram, BlackbirdProgram, PulseSequence, AnalogHamiltonianSimulation]]]): # noqa
                Single instance or list of circuits, annealing problems, pulse sequences,
                or photonics program to run on device.
            s3_destination_folder (Optional[S3DestinationFolder]): The S3 location to
                save the quantum tasks' results to. Default is `<default_bucket>/tasks` if evoked outside a
                Braket Job, `<Job Bucket>/jobs/<job name>/tasks` if evoked inside a Braket Job.
            shots (Optional[int]): The number of times to run the circuit or annealing problem.
                Default is 1000 for QPUs and 0 for simulators.
            max_parallel (Optional[int]): The maximum number of quantum tasks to run on AWS in parallel.
                Batch creation will fail if this value is greater than the maximum allowed
                concurrent quantum tasks on the device. Default: 10
            max_connections (int): The maximum number of connections in the Boto3 connection pool.
                Also the maximum number of thread pool workers for the batch. Default: 100
            poll_timeout_seconds (float): The polling timeout for `AwsQuantumTask.result()`,
                in seconds. Default: 5 days.
            poll_interval_seconds (float): The polling interval for `AwsQuantumTask.result()`,
                in seconds. Defaults to the ``getTaskPollIntervalMillis`` value specified in
                ``self.properties.service`` (divided by 1000) if provided, otherwise 1 second.
            inputs (Optional[Union[dict[str, float], list[dict[str, float]]]]): Inputs to be
                passed along with the IR. If the IR supports inputs, the inputs will be updated
                with this value. Default: {}.
            gate_definitions (Optional[dict[tuple[Gate, QubitSet], PulseSequence]]): A
                `dict[tuple[Gate, QubitSet], PulseSequence]]` for a user defined gate calibration.
                The calibration is defined for a particular `Gate` on a particular `QubitSet`
                and is represented by a `PulseSequence`. Default: None.
            reservation_arn (Optional[str]): The reservation ARN provided by Braket Direct
                to reserve exclusive usage for the device to run the quantum task on.
                Note: If you are creating tasks in a job that itself was created reservation ARN,
                those tasks do not need to be created with the reservation ARN.
                Default: None.

        Returns:
            AwsQuantumTaskBatch: A batch containing all of the quantum tasks run

        See Also:
            `braket.aws.aws_quantum_task_batch.AwsQuantumTaskBatch`
        """  # noqa E501
        if self._noise_model:
            task_specifications = [
                self._apply_noise_model_to_circuit(task_specification)
                for task_specification in task_specifications
            ]
        return AwsQuantumTaskBatch(
            AwsSession.copy_session(self._aws_session, max_connections=max_connections),
            self._arn,
            task_specifications,
            s3_destination_folder
            or (
                AwsSession.parse_s3_uri(os.environ.get("AMZN_BRAKET_TASK_RESULTS_S3_URI"))
                if "AMZN_BRAKET_TASK_RESULTS_S3_URI" in os.environ
                else None
            )
            or (self._aws_session.default_bucket(), "tasks"),
            shots if shots is not None else self._default_shots,
            max_parallel=max_parallel if max_parallel is not None else self._default_max_parallel,
            max_workers=max_connections,
            poll_timeout_seconds=poll_timeout_seconds,
            poll_interval_seconds=poll_interval_seconds or self._poll_interval_seconds,
            inputs=inputs,
            gate_definitions=gate_definitions,
            reservation_arn=reservation_arn,
            *aws_quantum_task_args,
            **aws_quantum_task_kwargs,
        )

    def refresh_metadata(self) -> None:
        """Refresh the `AwsDevice` object with the most recent Device metadata."""
        self._populate_properties(self._aws_session)

    def _get_session_and_initialize(self, session: AwsSession) -> AwsSession:
        device_region = AwsDevice.get_device_region(self._arn)
        return (
            self._get_regional_device_session(session)
            if device_region
            else self._get_non_regional_device_session(session)
        )

    def _get_regional_device_session(self, session: AwsSession) -> AwsSession:
        device_region = AwsDevice.get_device_region(self._arn)
        region_session = (
            session
            if session.region == device_region
            else AwsSession.copy_session(session, device_region)
        )
        try:
            self._populate_properties(region_session)
            return region_session
        except ClientError as e:
            raise (
                ValueError(f"'{self._arn}' not found")
                if e.response["Error"]["Code"] == "ResourceNotFoundException"
                else e
            )

    def _get_non_regional_device_session(self, session: AwsSession) -> AwsSession:
        current_region = session.region
        try:
            self._populate_properties(session)
            return session
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                if "qpu" not in self._arn:
                    raise ValueError(f"Simulator '{self._arn}' not found in '{current_region}'")
            else:
                raise e
        # Search remaining regions for QPU
        for region in frozenset(AwsDevice.REGIONS) - {current_region}:
            region_session = AwsSession.copy_session(session, region)
            try:
                self._populate_properties(region_session)
                return region_session
            except ClientError as e:
                if e.response["Error"]["Code"] != "ResourceNotFoundException":
                    raise e
        raise ValueError(f"QPU '{self._arn}' not found")

    def _populate_properties(self, session: AwsSession) -> None:
        metadata = session.get_device(self._arn)
        self._name = metadata.get("deviceName")
        self._status = metadata.get("deviceStatus")
        self._type = AwsDeviceType(metadata.get("deviceType"))
        self._provider_name = metadata.get("providerName")
        self._properties = BraketSchemaBase.parse_raw_schema(metadata.get("deviceCapabilities"))
        device_poll_interval = self._properties.service.getTaskPollIntervalMillis
        self._poll_interval_seconds = (
            device_poll_interval / 1000.0
            if device_poll_interval
            else AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL
        )
        self._topology_graph = None
        self._frames = None
        self._ports = None

    @property
    def type(self) -> str:
        """str: Return the device type"""
        return self._type

    @property
    def provider_name(self) -> str:
        """str: Return the provider name"""
        return self._provider_name

    @property
    def aws_session(self) -> AwsSession:
        return self._aws_session

    @property
    def arn(self) -> str:
        """str: Return the ARN of the device"""
        return self._arn

    @property
    def gate_calibrations(self) -> Optional[GateCalibrations]:
        """Calibration data for a QPU. Calibration data is shown for gates on particular gubits.
        If a QPU does not expose these calibrations, None is returned.

        Returns:
            Optional[GateCalibrations]: The calibration object. Returns `None` if the data
            is not present.
        """
        if not self._gate_calibrations:
            self._gate_calibrations = self.refresh_gate_calibrations()
        return self._gate_calibrations

    @property
    def is_available(self) -> bool:
        """Returns true if the device is currently available.

        Returns:
            bool: Return if the device is currently available.
        """
        if self.status != "ONLINE":
            return False

        is_available_result = False

        current_datetime_utc = datetime.utcnow()
        for execution_window in self.properties.service.executionWindows:
            weekday = current_datetime_utc.weekday()
            current_time_utc = current_datetime_utc.time().replace(microsecond=0)

            if current_time_utc < execution_window.windowEndHour < execution_window.windowStartHour:
                weekday = (weekday - 1) % 7

            matched_day = execution_window.executionDay == ExecutionDay.EVERYDAY
            matched_day = matched_day or (
                execution_window.executionDay == ExecutionDay.WEEKDAYS and weekday < 5
            )
            matched_day = matched_day or (
                execution_window.executionDay == ExecutionDay.WEEKENDS and weekday > 4
            )
            ordered_days = (
                ExecutionDay.MONDAY,
                ExecutionDay.TUESDAY,
                ExecutionDay.WEDNESDAY,
                ExecutionDay.THURSDAY,
                ExecutionDay.FRIDAY,
                ExecutionDay.SATURDAY,
                ExecutionDay.SUNDAY,
            )
            matched_day = matched_day or (
                execution_window.executionDay in ordered_days
                and ordered_days.index(execution_window.executionDay) == weekday
            )

            matched_time = (
                execution_window.windowStartHour < execution_window.windowEndHour
                and execution_window.windowStartHour
                <= current_time_utc
                <= execution_window.windowEndHour
            ) or (
                execution_window.windowEndHour < execution_window.windowStartHour
                and (
                    current_time_utc >= execution_window.windowStartHour
                    or current_time_utc <= execution_window.windowEndHour
                )
            )

            is_available_result = is_available_result or (matched_day and matched_time)

        return is_available_result

    @property
    # TODO: Add a link to the boto3 docs
    def properties(self) -> DeviceCapabilities:
        """DeviceCapabilities: Return the device properties

        Please see `braket.device_schema` in amazon-braket-schemas-python_

        .. _amazon-braket-schemas-python: https://github.com/aws/amazon-braket-schemas-python
        """
        return self._properties

    @property
    def topology_graph(self) -> DiGraph:
        """DiGraph: topology of device as a networkx `DiGraph` object.

        Examples:
            >>> import networkx as nx
            >>> device = AwsDevice("arn1")
            >>> nx.draw_kamada_kawai(device.topology_graph, with_labels=True, font_weight="bold")

            >>> topology_subgraph = device.topology_graph.subgraph(range(8))
            >>> nx.draw_kamada_kawai(topology_subgraph, with_labels=True, font_weight="bold")

            >>> print(device.topology_graph.edges)

        Returns:
            DiGraph: topology of QPU as a networkx `DiGraph` object. `None` if the topology
            is not available for the device.
        """
        if not self._topology_graph:
            self._topology_graph = self._construct_topology_graph()
        return self._topology_graph

    def _construct_topology_graph(self) -> DiGraph:
        """Construct topology graph. If no such metadata is available, return `None`.

        Returns:
            DiGraph: topology of QPU as a networkx `DiGraph` object.
        """
        if hasattr(self.properties, "paradigm") and isinstance(
            self.properties.paradigm, GateModelQpuParadigmProperties
        ):
            if self.properties.paradigm.connectivity.fullyConnected:
                return complete_graph(
                    int(self.properties.paradigm.qubitCount), create_using=DiGraph()
                )
            adjacency_lists = self.properties.paradigm.connectivity.connectivityGraph
            edges = []
            for item in adjacency_lists.items():
                i = item[0]
                edges.extend([(int(i), int(j)) for j in item[1]])
            return from_edgelist(edges, create_using=DiGraph())
        elif hasattr(self.properties, "provider") and isinstance(
            self.properties.provider, DwaveProviderProperties
        ):
            edges = self.properties.provider.couplers
            return from_edgelist(edges, create_using=DiGraph())
        else:
            return None

    @property
    def _default_shots(self) -> int:
        return (
            AwsDevice.DEFAULT_SHOTS_QPU if "qpu" in self.arn else AwsDevice.DEFAULT_SHOTS_SIMULATOR
        )

    @property
    def _default_max_parallel(self) -> int:
        return AwsDevice.DEFAULT_MAX_PARALLEL

    def __repr__(self):
        return f"Device('name': {self.name}, 'arn': {self.arn})"

    def __eq__(self, other: AwsDevice):
        if isinstance(other, AwsDevice):
            return self.arn == other.arn
        return NotImplemented

    @property
    def frames(self) -> dict[str, Frame]:
        """Returns a dict mapping frame ids to the frame objects for predefined frames
        for this device.
        """
        self._update_pulse_properties()
        return self._frames or {}

    @property
    def ports(self) -> dict[str, Port]:
        """Returns a dict mapping port ids to the port objects for predefined ports
        for this device.
        """
        self._update_pulse_properties()
        return self._ports or {}

    @staticmethod
    def get_devices(
        arns: Optional[list[str]] = None,
        names: Optional[list[str]] = None,
        types: Optional[list[AwsDeviceType]] = None,
        statuses: Optional[list[str]] = None,
        provider_names: Optional[list[str]] = None,
        order_by: str = "name",
        aws_session: Optional[AwsSession] = None,
    ) -> list[AwsDevice]:
        """Get devices based on filters and desired ordering. The result is the AND of
        all the filters `arns`, `names`, `types`, `statuses`, `provider_names`.

        Examples:
            >>> AwsDevice.get_devices(provider_names=['Rigetti'], statuses=['ONLINE'])
            >>> AwsDevice.get_devices(order_by='provider_name')
            >>> AwsDevice.get_devices(types=['SIMULATOR'])

        Args:
            arns (Optional[list[str]]): device ARN filter, default is `None`
            names (Optional[list[str]]): device name filter, default is `None`
            types (Optional[list[AwsDeviceType]]): device type filter, default is `None`
                QPUs will be searched for all regions and simulators will only be
                searched for the region of the current session.
            statuses (Optional[list[str]]): device status filter, default is `None`. When `None`
                is used, RETIRED devices will not be returned. To include RETIRED devices in
                the results, use a filter that includes "RETIRED" for this parameter.
            provider_names (Optional[list[str]]): provider name filter, default is `None`
            order_by (str): field to order result by, default is `name`.
                Accepted values are ['arn', 'name', 'type', 'provider_name', 'status']
            aws_session (Optional[AwsSession]): An AWS session object.
                Default is `None`.

        Raises:
            ValueError: order_by not in ['arn', 'name', 'type', 'provider_name', 'status']

        Returns:
            list[AwsDevice]: list of AWS devices
        """
        if order_by not in AwsDevice._GET_DEVICES_ORDER_BY_KEYS:
            raise ValueError(
                f"order_by '{order_by}' must be in {AwsDevice._GET_DEVICES_ORDER_BY_KEYS}"
            )
        types = frozenset(types or AwsDeviceType)
        aws_session = aws_session if aws_session else AwsSession()
        device_map = {}
        session_region = aws_session.boto_session.region_name
        search_regions = (
            (session_region,) if types == {AwsDeviceType.SIMULATOR} else AwsDevice.REGIONS
        )
        for region in search_regions:
            session_for_region = (
                aws_session
                if region == session_region
                else AwsSession.copy_session(aws_session, region)
            )
            # Simulators are only instantiated in the same region as the AWS session
            types_for_region = sorted(
                types if region == session_region else types - {AwsDeviceType.SIMULATOR}
            )
            try:
                region_device_arns = [
                    result["deviceArn"]
                    for result in session_for_region.search_devices(
                        arns=arns,
                        names=names,
                        types=types_for_region,
                        statuses=statuses,
                        provider_names=provider_names,
                    )
                ]
                device_map.update(
                    {
                        arn: AwsDevice(arn, session_for_region)
                        for arn in region_device_arns
                        if arn not in device_map
                    }
                )
            except ClientError as e:
                error_code = e.response["Error"]["Code"]
                warnings.warn(
                    f"{error_code}: Unable to search region '{region}' for devices."
                    " Please check your settings or try again later."
                    f" Continuing without devices in '{region}'.",
                    stacklevel=1,
                )

        devices = list(device_map.values())
        devices.sort(key=lambda x: getattr(x, order_by))
        return devices

    def _update_pulse_properties(self) -> None:
        if hasattr(self.properties, "pulse") and isinstance(
            self.properties.pulse, PulseDeviceActionProperties
        ):
            if self._ports is None:
                self._ports = {}
                port_data = self.properties.pulse.ports
                for port_id, port in port_data.items():
                    self._ports[port_id] = Port(
                        port_id=port_id, dt=port.dt, properties=json.loads(port.json())
                    )
            if self._frames is None:
                self._frames = {}
                frame_data = self.properties.pulse.frames
                if frame_data:
                    for frame_id, frame in frame_data.items():
                        self._frames[frame_id] = Frame(
                            frame_id=frame_id,
                            port=self._ports[frame.portId],
                            frequency=frame.frequency,
                            phase=frame.phase,
                            is_predefined=True,
                            properties=json.loads(frame.json()),
                        )

    @staticmethod
    def get_device_region(device_arn: str) -> str:
        """Gets the region from a device arn.

        Args:
            device_arn (str): The device ARN.

        Raises:
            ValueError: Raised if the ARN is not properly formatted

        Returns:
            str: the region of the ARN.
        """
        try:
            return device_arn.split(":")[3]
        except IndexError:
            raise ValueError(
                f"Device ARN is not a valid format: {device_arn}. For valid Braket ARNs, "
                "see 'https://docs.aws.amazon.com/braket/latest/developerguide/braket-devices.html'"
            )

    def queue_depth(self) -> QueueDepthInfo:
        """Task queue depth refers to the total number of quantum tasks currently waiting
        to run on a particular device.

        Returns:
            QueueDepthInfo: Instance of the QueueDepth class representing queue depth
            information for quantum tasks and hybrid jobs.
            Queue depth refers to the number of quantum tasks and hybrid jobs queued on a particular
            device. The normal tasks refers to the quantum tasks not submitted via Hybrid Jobs.
            Whereas, the priority tasks refers to the total number of quantum tasks waiting to run
            submitted through Amazon Braket Hybrid Jobs. These tasks run before the normal tasks.
            If the queue depth for normal or priority quantum tasks is greater than 4000, we display
            their respective queue depth as '>4000'. Similarly, for hybrid jobs if there are more
            than 1000 jobs queued on a device, display the hybrid jobs queue depth as '>1000'.
            Additionally, for QPUs if hybrid jobs queue depth is 0, we display information about
            priority and count of the running hybrid job.

        Example:
            Queue depth information for a running job.
            >>> device = AwsDevice(Device.Amazon.SV1)
            >>> print(device.queue_depth())
            QueueDepthInfo(quantum_tasks={<QueueType.NORMAL: 'Normal'>: '0',
            <QueueType.PRIORITY: 'Priority'>: '1'}, jobs='0 (1 prioritized job(s) running)')

            If more than 4000 quantum tasks queued on a device.
            >>> device = AwsDevice(Device.Amazon.DM1)
            >>> print(device.queue_depth())
            QueueDepthInfo(quantum_tasks={<QueueType.NORMAL: 'Normal'>: '>4000',
            <QueueType.PRIORITY: 'Priority'>: '2000'}, jobs='100')
        """
        metadata = self.aws_session.get_device(arn=self.arn)
        queue_metadata = metadata.get("deviceQueueInfo")
        queue_info = {}

        for response in queue_metadata:
            queue_name = response.get("queue")
            queue_priority = response.get("queuePriority")
            queue_size = response.get("queueSize")

            if queue_name == "QUANTUM_TASKS_QUEUE":
                priority_enum = QueueType(queue_priority)
                queue_info.setdefault("quantum_tasks", {})[priority_enum] = queue_size
            else:
                queue_info["jobs"] = queue_size

        return QueueDepthInfo(**queue_info)

    def refresh_gate_calibrations(self) -> Optional[GateCalibrations]:
        """Refreshes the gate calibration data upon request.

        If the device does not have calibration data, None is returned.

        Raises:
            URLError: If the URL provided returns a non 2xx response.

        Returns:
            Optional[GateCalibrations]: the calibration data for the device. None
            is returned if the device does not have a gate calibrations URL associated.
        """
        if (
            hasattr(self.properties, "pulse")
            and hasattr(self.properties.pulse, "nativeGateCalibrationsRef")
            and self.properties.pulse.nativeGateCalibrationsRef
        ):
            try:
                with urllib.request.urlopen(
                    self.properties.pulse.nativeGateCalibrationsRef.split("?")[0]
                ) as f:
                    json_calibration_data = self._parse_calibration_json(
                        json.loads(f.read().decode("utf-8"))
                    )
                    return GateCalibrations(json_calibration_data)
            except urllib.error.URLError:
                raise urllib.error.URLError(
                    f"Unable to reach {self.properties.pulse.nativeGateCalibrationsRef}"
                )
        else:
            return None

    def _parse_waveforms(self, waveforms_json: dict) -> dict:
        waveforms = {}
        for waveform in waveforms_json:
            parsed_waveform = _parse_waveform_from_calibration_schema(waveforms_json[waveform])
            waveforms[parsed_waveform.id] = parsed_waveform
        return waveforms

    def _parse_pulse_sequence(
        self, calibration: dict, waveforms: dict[ArbitraryWaveform]
    ) -> PulseSequence:
        return PulseSequence._parse_from_calibration_schema(calibration, waveforms, self.frames)

    def _parse_calibration_json(
        self, calibration_data: dict
    ) -> dict[tuple[Gate, QubitSet], PulseSequence]:
        """Takes the json string from the device calibration URL and returns a structured dictionary of
        corresponding `dict[tuple[Gate, QubitSet], PulseSequence]` to represent the calibration data.

        Args:
            calibration_data (dict): The data to be parsed. Based on
                https://github.com/aws/amazon-braket-schemas-python/blob/main/src/braket/device_schema/pulse/native_gate_calibrations_v1.py.

        Returns:
            dict[tuple[Gate, QubitSet], PulseSequence]: The
            structured data based on a mapping of `tuple[Gate, Qubit]` to its calibration repesented as a
            `PulseSequence`.

        """  # noqa: E501
        waveforms = self._parse_waveforms(calibration_data["waveforms"])
        parsed_calibration_data = {}
        for qubit_node in calibration_data["gates"]:
            qubit = calibration_data["gates"][qubit_node]
            for gate_node in qubit:
                for gate in qubit[gate_node]:
                    gate_capitalized = getattr(
                        self,
                        f"_{self.provider_name.upper()}_GATES_TO_BRAKET",
                        {},
                    ).get(gate_node.capitalize(), gate_node.capitalize())
                    gate_obj = (
                        getattr(importlib.import_module("braket.circuits.gates"), gate_capitalized)
                        if gate_capitalized is not None
                        else None
                    )
                    qubits = QubitSet([int(x) for x in gate["qubits"]])
                    if gate_obj is None:
                        # We drop out gates that are not implemented in the BDK
                        continue

                    argument = None
                    if gate["arguments"]:
                        argument = (
                            float(gate["arguments"][0])
                            if _is_float(gate["arguments"][0])
                            else FreeParameter(gate["arguments"][0])
                        )
                    gate_qubit_key = (
                        (gate_obj(argument), qubits) if argument else (gate_obj(), qubits)
                    )
                    gate_qubit_pulse = self._parse_pulse_sequence(gate["calibrations"], waveforms)
                    parsed_calibration_data[gate_qubit_key] = gate_qubit_pulse

        return parsed_calibration_data



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys


def running_in_jupyter() -> bool:
    """Determine if running within Jupyter.

    Inspired by https://github.com/ipython/ipython/issues/11694

    Returns:
        bool: True if running in Jupyter, else False.
    """
    in_ipython = False
    in_ipython_kernel = False

    # if IPython hasn't been imported, there's nothing to check
    if "IPython" in sys.modules:
        get_ipython = sys.modules["IPython"].__dict__["get_ipython"]

        ip = get_ipython()
        in_ipython = ip is not None

    if in_ipython:
        in_ipython_kernel = getattr(ip, "kernel", None) is not None

    return in_ipython_kernel



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Any

import braket.ir.jaqcd as ir
from braket.circuits.compiler_directive import CompilerDirective


class StartVerbatimBox(CompilerDirective):
    """Prevents the compiler from modifying any ensuing instructions
    until the appearance of a corresponding ``EndVerbatimBox``.
    """

    def __init__(self):
        super().__init__(["StartVerbatim"])

    def counterpart(self) -> CompilerDirective:
        return EndVerbatimBox()

    def _to_jaqcd(self, *args, **kwargs) -> Any:
        return ir.StartVerbatimBox.construct()

    def _to_openqasm(self) -> str:
        return "#pragma braket verbatim\nbox{"


class EndVerbatimBox(CompilerDirective):
    """Marks the end of a portion of code following a StartVerbatimBox that prevents the enclosed
    instructions from being modified by the compiler.
    """

    def __init__(self):
        super().__init__(["EndVerbatim"])

    def counterpart(self) -> CompilerDirective:
        return StartVerbatimBox()

    def _to_jaqcd(self, *args, **kwargs) -> Any:
        return ir.EndVerbatimBox.construct()

    def _to_openqasm(self) -> str:
        return "}"



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Operator
from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.serialization import IRType


@pytest.fixture
def ascii_symbols():
    return ["foo"]


@pytest.fixture
def compiler_directive(ascii_symbols):
    return CompilerDirective(ascii_symbols=ascii_symbols)


def test_is_operator(compiler_directive):
    assert isinstance(compiler_directive, Operator)


def test_ascii_symbols(compiler_directive, ascii_symbols):
    assert compiler_directive.ascii_symbols == tuple(ascii_symbols)


def test_none_ascii():
    with pytest.raises(ValueError):
        CompilerDirective(ascii_symbols=None)


def test_name(compiler_directive):
    expected = compiler_directive.__class__.__name__
    assert compiler_directive.name == expected


def test_counterpart_not_implemented_by_default(compiler_directive):
    with pytest.raises(NotImplementedError):
        compiler_directive.counterpart()


def test_str(compiler_directive):
    assert str(compiler_directive) == compiler_directive.name


@pytest.mark.parametrize(
    "ir_type, serialization_properties, expected_exception, expected_message",
    [
        (IRType.JAQCD, None, NotImplementedError, "to_jaqcd has not been implemented yet."),
        (IRType.OPENQASM, None, NotImplementedError, "to_openqasm has not been implemented yet."),
        ("invalid-ir-type", None, ValueError, "Supplied ir_type invalid-ir-type is not supported."),
    ],
)
def test_compiler_directive_to_ir(
    ir_type, serialization_properties, expected_exception, expected_message, compiler_directive
):
    with pytest.raises(expected_exception) as exc:
        compiler_directive.to_ir(0, ir_type, serialization_properties=serialization_properties)
    assert exc.value.args[0] == expected_message



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.circuits import Circuit, Observable
from braket.devices import LocalSimulator

device = LocalSimulator()

print("Example for shots=0")
# Result types can be requested in the circuit
# Example of result types for shots=0
bell = (
    Circuit()
    .h(0)
    .cnot(0, 1)
    .probability(target=[0])
    .expectation(observable=Observable.Z(), target=[1])
    .amplitude(state=["00"])
    .state_vector()
)

# State vector and amplitude can only be requested when shots=0 for a simulator
# When shots=0 for a simulator, probability, expectation, variance are the exact values,
# not calculated from measurements
# Users cannot request Sample as a result when shots=0
result = device.run(bell).result()
print("Marginal probability for target 0 in computational basis:", result.values[0])
print("Expectation of target 1 in the computational basis:", result.values[1])
print("Amplitude of state 00:", result.values[2])
print("State vector:", result.values[3])

print("\nExample for shots>0")
# Example of result types for shots > 0
bell = (
    Circuit()
    .h(0)
    .cnot(0, 1)
    .expectation(observable=Observable.Y() @ Observable.X(), target=[0, 1])
    .variance(observable=Observable.Y() @ Observable.X(), target=[0, 1])
    .sample(observable=Observable.Y() @ Observable.X(), target=[0, 1])
)

# When shots>0 for a simulator, probability, expectation, variance are calculated from measurements
# Users can request sample as a result when shots > 0
result = device.run(bell, shots=100).result()
print("Expectation of target 0, 1 in the basis of Pauli-Y @ Pauli-X:", result.values[0])
print("Variance of target 0, 1 in the basis of Pauli-Y @ Pauli-X:", result.values[1])
print("Samples of target 0, 1 in the basis of Pauli-Y @ Pauli-X:", result.values[2])

# Probability, sample, expectation, and variance are also supported for QPU devices
